#    This file is part of NEORL.

#    Copyright (c) 2021 Exelon Corporation and MIT Nuclear Science and Engineering
#    NEORL is free software: you can redistribute it and/or modify
#    it under the terms of the MIT LICENSE

#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#    SOFTWARE.

#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jun 29 15:36:46 2020

@author: alyssawang
"""

#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Mar  4 11:51:22 2020

@author: majdi
"""

import numpy as np
import pandas as pd
import os
import random
import itertools
import sys, copy, shutil
import time
import pickle
import subprocess
from multiprocessing import Pool
from skopt import Optimizer

from skopt import gp_minimize, dump, load
from skopt.space import Real, Integer, Categorical
from skopt.plots import plot_convergence
from skopt.utils import use_named_args
import random
import numpy 
import matplotlib.pyplot as plt

class BAYESTUNE:
    
    """
    A class to parse neorl input template and construct cases for bayesian optimization for hyperparameters

    inputs: 
    The template input file
    Class object from PARSER.py, featuring user input for TUNE
    neorl logo
    """
    
    def __init__(self, inputfile, tuneblock, logo):
        self.logo=logo
        self.inputfile=inputfile
        self.tuneblock=tuneblock
        self.n_last_episodes=int(self.tuneblock["n_last_episodes"])
        self.ncores=int(self.tuneblock["ncores"])
        self.ncases=int(self.tuneblock["ncases"])
                
        #---------------------------------------
        # define bayesian optimization parameters
        #---------------------------------------
        self.caseindex=0 # for creating case directory purposes 
        
        #-------------------------------
        # construct results directory
        #-------------------------------
        if os.path.exists('./tunecases/'):
            shutil.rmtree('./tunecases/')
            os.makedirs('./tunecases/', exist_ok=True)
        else:
            os.makedirs('./tunecases/', exist_ok=True)
        self.csvlogger='tune.csv'
        self.tunesummary='tunesummary.txt'
        
        #---------------------------------
        # parse the input template
        #---------------------------------
        with open (self.inputfile, 'r') as input_file_text:
            self.template=input_file_text.readlines()
            
        first=0; last=0
        for i in range (len(self.template)):
            if ('READ TUNE' in self.template[i]):
                first=i
            if ('END TUNE' in self.template[i]):
                last=i
        if first == 0 and last==0:
            raise ('TUNE card cannot be found')
        
        del self.template[first : last+1]  
        self.template="".join(self.template)  
        
                     
    def tune_count(self):
        
        """
        1- This function uses self.tuneblock, parse it, infer all parameters to be tuned and thier distribution
        2- This function instantiates the search dimensions for the hyperparameters to be tuned
        """
        
        self.param_dict={}
        for item in self.tuneblock:
            if '{' in item and '}' in item and item[0] != '#':
                #-----------------------------------------------------
                # check the existence of the name in the template
                #-----------------------------------------------------
                if item not in self.template:
                    raise ValueError('parameter {} in TUNE block cannot be found in any other block, e.g. DQN, GA, PPO, etc.'.format(item)) 

                item_lst=self.tuneblock[item].split(",")
                item_lst=[item.strip() for item in item_lst] # get rid of white spaces in the splitted values
                #-------------------------------------------------------
                # check if a uniform distribution of floats is identified
                #-------------------------------------------------------
                try:
                    if "float" in item_lst:
                        item_lst[0]=float(item_lst[0])
                        item_lst[1]=float(item_lst[1])
                        print ('-- debug: parameter {} has uniform distribution of type --float-- between {} and {}'.format(item,item_lst[0],item_lst[1]))
                    elif "u" in item_lst:
                        item_lst[0]=float(item_lst[0])
                        item_lst[1]=float(item_lst[1])
                        print ('-- debug: parameter {} has uniform distribution of type --float-- between {} and {}'.format(item,item_lst[0],item_lst[1]))
                except:
                    raise Exception ('--error: TUNE cannot construct the user-given uniform distribution of --floats-- for {} according to (low, high, u) syntax'.format(item))
               
                #----------------------------------------------------
                # check if a random integer distribution is identified
                #----------------------------------------------------
                try:
                    if "int" in item_lst:
                        item_lst[0]=int(item_lst[0])
                        item_lst[1]=int(item_lst[1])
                        print ('-- debug: parameter {} has uniform distribution of type --int-- between {} and {}'.format(item,item_lst[0],item_lst[1]))
                    elif "randint" in item_lst:
                        item_lst[0]=int(item_lst[0])
                        item_lst[1]=int(item_lst[1])
                        print ('-- debug: parameter {} has uniform distribution of type --int-- between {} and {}'.format(item,item_lst[0],item_lst[1]))
                except:
                    raise Exception ('--error: TUNE cannot construct the user-given uniform distribution of --int-- for {} according to (low, high, u) syntax'.format(item))
        
               #-----------------------------------------------------
               # check if a grid is identified
               #-----------------------------------------------------
                try:
                    if "grid" in item_lst:
                        element_lst=[]
                        for element in item_lst:
                            # check if it is an integer
                            not_int=0
                            try:
                                element_lst.append(int(element.strip()))
                            except Exception:
                                not_int=1
                            
                            # else check if the elment is float
                            if not_int:
                                try:
                                    element_lst.append(float(element.strip()))
                                # else consider it a string
                                except Exception:
                                    element_lst.append(str(element.strip()))
                                    
                        item_lst=element_lst
                        print ('-- debug: parameter {} has grid type with values {}'.format(item,item_lst))
                except:
                    raise Exception ('--error: TUNE cannot construct the user-given grid for {} according to the comma-seperated syntax'.format(item))

                self.param_dict[item]=item_lst # Save the final parsed list for parameter {XXX}                    
                
        
        #------------------------------------------------------------------
        # Instantiate the search space dimensions for bayesian optimization
        #------------------------------------------------------------------
        #random.seed(64)
                
        self.dimensions = []            
        self.initialparams = [ [] for _ in range(self.ncores)]
          
        for key in list(self.param_dict.keys()):
            
            if ('float' in self.param_dict[key]) or ('u' in self.param_dict[key]):  
                # set dimension
                self.dimensions.append(Real(low=self.param_dict[key][0], high=self.param_dict[key][1]))
                # add to initial params
                for i in range(self.ncores):
                    self.initialparams[i].append(random.uniform(self.param_dict[key][0], self.param_dict[key][1]))
                
            if ('int' in self.param_dict[key]) or ('randint' in self.param_dict[key]):
                # set dimension
                self.dimensions.append(Integer(low=self.param_dict[key][0], high=self.param_dict[key][1]))
                # add to initial params
                for i in range(self.ncores):
                    self.initialparams[i].append(random.randint(self.param_dict[key][0], self.param_dict[key][1]))

            if 'grid' in self.param_dict[key]:
                self.real_grid=list(self.param_dict[key])
                self.real_grid.remove('grid') # get rid of the 'grid' to avoid sampling it
                # set dimension
                self.dimensions.append(Categorical(categories=self.real_grid))
                # add to initial params
                for i in range(self.ncores):
                    self.initialparams[i].append(random.sample(self.real_grid, 1)[0])
                
    def gen_cases(self, x=0):
        
        """
        This function infers neorl.py path 
        """
        
        self.tune_count()
        self.param_names=list(self.param_dict.keys())
        # Find neorl path
        #self.here=os.path.dirname(os.path.abspath(__file__))
        #self.neorl_path=self.here.replace('src/tune','neorl.py') #try to infer neorl.py internally to call neorl inside or neorl
        #self.python_path=self.here.replace('neorl/src/tune','anaconda3/bin/python3') #try to infer python3 path to call neorl inside or neorl
        self.neorl_path=sys.argv[0]
        self.python_path=sys.executable
        print('--debug: NEORLPATH=', self.neorl_path)
        print('--debug: PYTHONPATH=', self.python_path)

    def evalX(self, params):
        
        """
        This function evaluates objective function value, calls self.case_object(params) for evaluation
        Inputs: 
            - Single input point
        Outputs: 
            - Objective function value at the input point  
        """
        
        return self.case_object(params)
        
    def case_object(self,CASEPARAMS): 
        
        """
        This function sets up a case object for a single input point during bayesian optimization
        Inputs:
            - Single input point
        Outputs:
            - Negative mean reward for the input point (to be minimized to find maximal mean reward)
        """
        
        try:
            #--------------------------------------------------------------------------------------------------------------
            # Prepares directories and files for one case
            self.param_names=list(self.param_dict.keys())
            i = self.caseindex
            os.makedirs('./tunecases/case{}'.format(i+1), exist_ok=True)
            self.new_template=copy.deepcopy(self.template)
            for j in range (len(self.param_names)):
                self.new_template=self.new_template.replace(str(self.param_names[j]), str(CASEPARAMS[j]))
            
            filename='./tunecases/case{}/case{}.inp'.format(i+1, i+1)
            with open (filename, 'w') as fout:
                fout.writelines(self.new_template)
             
            # copy external files into the new directory, if extfiles card exists
            if 'extfiles' in self.tuneblock:
                if self.tuneblock['extfiles']:
                    print('--debug: external files are identified, copying them into each case directory')
                    for item in self.tuneblock['extfiles']:
                        os.system('cp -r {} ./tunecases/case{}/'.format(item, i+1))
            
            
            casenum = i+1
            print('--------------------------------------------------')
            print('Running TUNE Case {}/{}: {}'.format(i+1, self.ncases, CASEPARAMS))
            subprocess.call([self.python_path, self.neorl_path, '-i', 'case{}.inp'.format(casenum)], cwd='./tunecases/case{}/'.format(casenum))  # this exceutes neorl for this case.inp
            print('--------------------------------------------------')
            
            #--------------------------------------------------------------------------------------------------------------
            # Try to infer the _out.csv file in the directory since only one method is allowed
            csvfile=[f for f in os.listdir('./tunecases/case{}/case{}_log/'.format(casenum, casenum)) if f.endswith('_out.csv')]
            if len(csvfile) > 1:
                raise Exception ('multiple *_out.csv files can be found in the logger of TUNE, only one is allowed')
            #--------------------------------------------------------------------------------------------------------------
            
            reward_lst=pd.read_csv('./tunecases/case{}/case{}_log/{}'.format(casenum, casenum, csvfile[0]), usecols=['reward']).values
            mean_reward=np.mean(reward_lst[-self.n_last_episodes:])
            max_reward=np.max(reward_lst)
            
            with open (self.csvlogger, 'a') as fout:
                fout.write(str(casenum) +',')
                [fout.write(str(item) + ',') for item in CASEPARAMS]
                fout.write(str(mean_reward) + ',' + str(max_reward) + '\n')
            
            self.caseindex+=1 
            return -mean_reward
        
        except:
            print('--error: case{}.inp failed during execution'.format(casenum))
            
            return 'case{}.inp:failed'.format(casenum)
        
    
    def run_gp_minimize(self, initialx):
        
        """
        This function calls gp_minimize of a list of initial input points
        Inputs:
            - List of initial input points
        Outputs: 
            - The optimization result returned as a OptimizeResult scipy object 
        """
        
        self.caseindex=initialx[-1]
        initialx=initialx[:-1]
               
        return gp_minimize(func=self.evalX, 
                            dimensions=self.dimensions,
                            acq_func='EI',
                            n_calls=self.n_calls,
                            x0=initialx,
                            random_state=self.caseindex)
    
    def run_cases(self):
        
        """
        This function runs bayesian optimization and collects their stats
        """
        
        with open (self.csvlogger, 'w') as fout:
            fout.write('caseid, ')
            [fout.write(item + ',') for item in self.param_names]
            fout.write('mean_reward,max_reward\n')
        
        #random.seed(64)
        
        if self.ncores==1: 
            #------------------------------
            # running process over one core
            #------------------------------
            if self.ncases < 11:
                self.ncases = 11
            
            self.results = gp_minimize(func=self.evalX,
                                        dimensions=self.dimensions,
                                        acq_func='EI',
                                        n_calls=self.ncases,
                                        x0=self.initialparams)
        else:
            #---------------------------------------------
            # running process in parallel over multi cores
            #---------------------------------------------
            if int(self.ncases/self.ncores) < 11:
                self.n_calls = 11
            else:
                self.n_calls = int(self.ncases/self.ncores)
            index=0
            for params in self.initialparams:
                params.append(index)
                index+=self.n_calls
            
            p=Pool(self.ncores)
            self.results = p.map(self.run_gp_minimize, self.initialparams)
            p.close()
            p.join()
        
                    
        csvdata=pd.read_csv('tune.csv')
        asc_data=csvdata.sort_values(by=['caseid'],ascending=True)
        des_data=csvdata.sort_values(by=['mean_reward'],ascending=False)
        des_data2=csvdata.sort_values(by=['max_reward'],ascending=False)
        asc_data.to_csv('tune.csv', index=False)
    
        mean = np.mean(des_data.iloc[:,4:5])
        totalmean=mean.tolist()[0]
        try:
            failed_cases=len([print ('failed') for item in self.results["func_vals"] if isinstance(item, str)])
        except:
            failed_cases='NA'
        
        print ('%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%')
        print('Mean Rewards for all cases=', totalmean)
        print ('%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%')
        print ('All TUNE CASES ARE COMPLETED')
        print ('%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%')
        print('--debug: Check tunesummary.txt file for best hyperparameters found')
        print('--debug: Check tune.csv file for complete csv logger of all cases results')
        print('--debug: Check tunecases directory for case-by-case detailed results')
        
        with open ('tunesummary.txt', 'w') as fout:
            
            #fout.write(self.logo)
            fout.write('*****************************************************\n')
            fout.write('Summary for the TUNE case \n')
            fout.write('*****************************************************\n')
            fout.write('Number of cases evaluated: {} \n'.format(self.ncases))
            fout.write('Number of failed cases: {} \n'.format(failed_cases))
            fout.write('Parameter names: {} \n'.format(self.param_names))
            fout.write('Parameter values: {} \n '.format(self.param_dict))
            fout.write ('--------------------------------------------------------------------------------------\n')
            if des_data.shape[0] < 20:
                top=des_data.shape[0]
                fout.write ('Top {} hyperparameter configurations ranked according to MEAN reward \n'.format(top))
                fout.write(des_data.iloc[:top].to_string(index=False))
            else:
                top=20
                fout.write ('Top {} hyperparameter configurations ranked according to MEAN reward \n'.format(top))
                fout.write(des_data.iloc[:top].to_string(index=False))
            fout.write ('\n')
            fout.write ('--------------------------------------------------------------------------------------\n')
            if des_data2.shape[0] < 20:
                top=des_data2.shape[0]
                fout.write ('Top {} hyperparameter configurations ranked according to MAX reward \n'.format(top))
                fout.write(des_data2.iloc[:top].to_string(index=False))
            else:
                top=20
                fout.write ('Top {} hyperparameter configurations ranked according to MAX reward \n'.format(top))
                fout.write(des_data2.iloc[:top].to_string(index=False))


#    This file is part of NEORL.

#    Copyright (c) 2021 Exelon Corporation and MIT Nuclear Science and Engineering
#    NEORL is free software: you can redistribute it and/or modify
#    it under the terms of the MIT LICENSE

#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#    SOFTWARE.

#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#"""
#Created on Wed Mar  4 11:51:22 2020
#
#@author: majdi
#"""

import logging
import numpy as np
import pandas as pd
import itertools
from multiprocessing import Pool
import joblib
import csv

class GRIDTUNE:
    """
    A module for grid search for hyperparameter tuning

    :param param_grid: (dict) the grid (list of possible values) for each hyperparameter provided in a dictionary form. Example: {'x1': [40, 50, 60, 80, 100], 'x2': [0.2, 0.4, 0.8], 'x3': ['blend', 'cx2point']}
    :param fit: (function) the self-defined fitness function that includes the hyperparameters as input and algorithm score as output
    """
    def __init__(self, param_grid, fit):
        self.param_grid=param_grid
        self.fit=fit
        self.full_grid()

    def full_grid(self):
        #This function builds the full multi-dimensional grid

        self.param_lst=[self.param_grid[item] for item in self.param_grid]
        self.param_names=[item for item in self.param_grid]
        #count all possible combinations      
        self.hyperparameter_cases = list(itertools.product(*self.param_lst)) # * here helps passing list of lists to product function 
                                                                    #   without need to know the size of parameters beforehand
         
    def worker(self,x):
        #This function setup a case object to pass to the Parallel pool

        caseid=x[0]
        param_vals=x[1]
        
        #form the dictionary for this case
        case_dict={}
        case_dict['id']=caseid
        assert len(param_vals) == len(self.param_names), '--error: it seems the length of the param_names ({}) and param_values ({}) are not equal, cannot proceed'.format(len(self.param_names), len(x))
        for name, val in zip(self.param_names, param_vals):
            case_dict[name]=val
        
        try:
            obj=self.fit(*param_vals)
            case_dict['score']=obj
            if self.verbose:
                print('-------------------------------------------------------------------------------------------')
                print('TUNE Case {}/{} is completed'.format(caseid, len(self.hyperparameter_cases), case_dict))
                print(case_dict)
                print('-------------------------------------------------------------------------------------------')
            
            if self.csvlogger:
                with open (self.csvlogger, 'a') as csvfile:
                    csvwriter = csv.writer(csvfile, delimiter=',', quoting=csv.QUOTE_MINIMAL, lineterminator = '\n')
                    csvwriter.writerow([case_dict[item] for item in case_dict])
            
            return obj
        
        except Exception as e:
            print(e)
            logging.exception("message")
            print('--error: case {} failed during execution'.format(caseid))
            print('--error: {} failed'.format(case_dict))
            
            return 'case{}:failed'.format(caseid)
        
        
    def tune(self, ncores=1, csvname=None, verbose=True):
        """
        This function starts the tuning process with specified number of processors
    
        :param ncores: (int) number of parallel processors (see the **Notes** section below for an important note about parallel execution)
        :param csvname: (str) the name of the csv file name to save the tuning results (useful for expensive cases as the csv file is updated directly after the case is done)
        :param verbose: (bool) whether to print updates to the screen or not
        """
        self.ncores=ncores
        self.csvlogger=csvname
        self.verbose=verbose

        if self.verbose:
            print('***************************************************************')
            print('****************Grid Search is Running*************************')
            print('***************************************************************')
            
            if self.ncores > 1:
                print('--- Running in parallel with {} cores'.format(self.ncores))
                
        if self.csvlogger:
            headers=['id']  + self.param_names + ['score']
            with open (self.csvlogger, 'w') as csvfile:
                csvwriter = csv.writer(csvfile, delimiter=',', quoting=csv.QUOTE_MINIMAL, lineterminator = '\n')
                csvwriter.writerow(headers)
                
        core_lst=[]
        for i in range (len(self.hyperparameter_cases)):
            core_lst.append([i+1, self.hyperparameter_cases[i]])
        
        if self.ncores > 1:
            #p=Pool(self.ncores)
            #results = p.map(self.worker, core_lst)
            #p.close()
            #p.join()
            
            with joblib.Parallel(n_jobs=self.ncores) as parallel:
                results=parallel(joblib.delayed(self.worker)(item) for item in core_lst)
                
        else:
            results=[]
            for item in core_lst:
                results.append(self.worker(item))

        gridres = pd.DataFrame(self.hyperparameter_cases, columns=self.param_names)
        gridres.index += 1
        gridres['score'] = results
        #gridres = gridres.sort_values(['score'], axis='index', ascending=False)     
        
        return gridres
        



import os

from gym.wrappers.monitoring import video_recorder

from neorl.rl.baselines.shared import logger
from neorl.rl.baselines.shared.vec_env.base_vec_env import VecEnvWrapper
from neorl.rl.baselines.shared.vec_env.dummy_vec_env import DummyVecEnv
from neorl.rl.baselines.shared.vec_env.subproc_vec_env import SubprocVecEnv
from neorl.rl.baselines.shared.vec_env.vec_frame_stack import VecFrameStack
from neorl.rl.baselines.shared.vec_env.vec_normalize import VecNormalize


class VecVideoRecorder(VecEnvWrapper):
    """
    Wraps a VecEnv or VecEnvWrapper object to record rendered image as mp4 video.
    It requires ffmpeg or avconv to be installed on the machine.

    :param venv: (VecEnv or VecEnvWrapper)
    :param video_folder: (str) Where to save videos
    :param record_video_trigger: (func) Function that defines when to start recording.
                                        The function takes the current number of step,
                                        and returns whether we should start recording or not.
    :param video_length: (int)  Length of recorded videos
    :param name_prefix: (str) Prefix to the video name
    """

    def __init__(self, venv, video_folder, record_video_trigger,
                 video_length=200, name_prefix='rl-video'):

        VecEnvWrapper.__init__(self, venv)

        self.env = venv
        # Temp variable to retrieve metadata
        temp_env = venv

        # Unwrap to retrieve metadata dict
        # that will be used by gym recorder
        while isinstance(temp_env, VecNormalize) or isinstance(temp_env, VecFrameStack):
            temp_env = temp_env.venv

        if isinstance(temp_env, DummyVecEnv) or isinstance(temp_env, SubprocVecEnv):
            metadata = temp_env.get_attr('metadata')[0]
        else:
            metadata = temp_env.metadata

        self.env.metadata = metadata

        self.record_video_trigger = record_video_trigger
        self.video_recorder = None

        self.video_folder = os.path.abspath(video_folder)
        # Create output folder if needed
        os.makedirs(self.video_folder, exist_ok=True)

        self.name_prefix = name_prefix
        self.step_id = 0
        self.video_length = video_length

        self.recording = False
        self.recorded_frames = 0

    def reset(self):
        obs = self.venv.reset()
        self.start_video_recorder()
        return obs

    def start_video_recorder(self):
        self.close_video_recorder()

        video_name = '{}-step-{}-to-step-{}'.format(self.name_prefix, self.step_id,
                                                    self.step_id + self.video_length)
        base_path = os.path.join(self.video_folder, video_name)
        self.video_recorder = video_recorder.VideoRecorder(
                env=self.env,
                base_path=base_path,
                metadata={'step_id': self.step_id}
                )

        self.video_recorder.capture_frame()
        self.recorded_frames = 1
        self.recording = True

    def _video_enabled(self):
        return self.record_video_trigger(self.step_id)

    def step_wait(self):
        obs, rews, dones, infos = self.venv.step_wait()

        self.step_id += 1
        if self.recording:
            self.video_recorder.capture_frame()
            self.recorded_frames += 1
            if self.recorded_frames > self.video_length:
                logger.info("Saving video to ", self.video_recorder.path)
                self.close_video_recorder()
        elif self._video_enabled():
            self.start_video_recorder()

        return obs, rews, dones, infos

    def close_video_recorder(self):
        if self.recording:
            self.video_recorder.close()
        self.recording = False
        self.recorded_frames = 1

    def close(self):
        VecEnvWrapper.close(self)
        self.close_video_recorder()

    def __del__(self):
        self.close()



#    This file is part of NEORL.

#    Copyright (c) 2021 Exelon Corporation and MIT Nuclear Science and Engineering
#    NEORL is free software: you can redistribute it and/or modify
#    it under the terms of the MIT LICENSE

#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#    SOFTWARE.

# -*- coding: utf-8 -*-
#"""
#Created on Fri Jun 18 19:45:06 2021
#
#@author: Devin Seyler and Majdi Radaideh
#"""

import random
import numpy as np
import math
import joblib
from neorl.evolu.discrete import mutate_discrete, encode_grid_to_discrete 
from neorl.evolu.discrete import decode_discrete_to_grid, encode_grid_indv_to_discrete
from neorl.utils.seeding import set_neorl_seed
from neorl.utils.tools import get_population, check_mixed_individual

#Main reference of the BAT algorithm:
#Xie, J., Zhou, Y., & Chen, H. (2013). A novel bat algorithm based on 
#differential operator and Levy flights trajectory. 
#Computational intelligence and neuroscience, 2013.

class BAT(object):
    """
    BAT Algorithm
    
    :param mode: (str) problem type, either ``min`` for minimization problem or ``max`` for maximization
    :param bounds: (dict) input parameter type and lower/upper bounds in dictionary form. Example: ``bounds={'x1': ['float', 0.1, 0.8], 'x2': ['float', 2.2, 6.2]}``
    :param fit: (function) the fitness function 
    :param nbats: (int): number of bats in the population
    :param fmin: (float): minimum value of the pulse frequency
    :param fmax: (float): maximum value of the pulse frequency
    :param A: (float) initial value of the loudness rate
    :param r0: (float) asymptotic value of the pulse rate 
    :param alpha: (float) decay factor of loudness ``A``, i.e. A approaches 0 by the end of evolution if ``alpha < 1``
    :param gamma: (float) exponential factor of the pulse rate ``r``, i.e. ``r`` increases abruptly at the beginning and then converges to ``r0`` by the end of evolution
    :param levy: (bool): a flag to activate Levy flight steps of the bat to increase bat diversity
    :param int_transform: (str): method of handling int/discrete variables, choose from: ``nearest_int``, ``sigmoid``, ``minmax``.
    :param ncores: (int) number of parallel processors (must be ``<= nbats``)
    :param seed: (int) random seed for sampling
    """
    def __init__(self, mode, bounds, fit, nbats=50, fmin=0, 
                 fmax=1, A=0.5, r0=0.5, alpha=1.0, gamma=0.9, 
                 levy='False', int_transform='nearest_int', ncores=1, seed=None):
        
        set_neorl_seed(seed)
        
        assert ncores <= nbats, '--error: ncores ({}) must be less than or equal to nbats ({})'.format(ncores, nbats)
        assert nbats >= 5, '--error: number of bats must be more than 5 for this algorithm'
        
        #--mir
        self.mode=mode
        if mode == 'min':
            self.fit=fit
        elif mode == 'max':
            def fitness_wrapper(*args, **kwargs):
                return -fit(*args, **kwargs) 
            self.fit=fitness_wrapper
        else:
            raise ValueError('--error: The mode entered by user is invalid, use either `min` or `max`')
        
        self.int_transform=int_transform
        if int_transform not in ["nearest_int", "sigmoid", "minmax"]:
            raise ValueError('--error: int_transform entered by user is invalid, must be `nearest_int`, `sigmoid`, or `minmax`')
            
        self.bounds=bounds
        self.ncores = ncores
        self.nbats=nbats
        
        self.fmax=fmax
        self.fmin=fmin
        self.A=A
        self.alpha=alpha
        self.gamma=gamma
        self.r0=r0
        self.levy_flight=levy
        
        #infer variable types 
        self.var_type = np.array([bounds[item][0] for item in bounds])
        
        #mir-grid
        if "grid" in self.var_type:
            self.grid_flag=True
            self.orig_bounds=bounds  #keep original bounds for decoding
            #print('--debug: grid parameter type is found in the space')
            self.bounds, self.bounds_map=encode_grid_to_discrete(self.bounds) #encoding grid to int
            #define var_types again by converting grid to int
            self.var_type = np.array([self.bounds[item][0] for item in self.bounds])
        else:
            self.grid_flag=False
            self.bounds = bounds
            self.orig_bounds=bounds
        
        self.dim = len(bounds)
        self.lb=np.array([self.bounds[item][1] for item in self.bounds])
        self.ub=np.array([self.bounds[item][2] for item in self.bounds])

    def init_sample(self, bounds):
        #sample initializer
        indv=[]
        for key in bounds:
            if bounds[key][0] == 'int':
                indv.append(random.randint(bounds[key][1], bounds[key][2]))
            elif bounds[key][0] == 'float':
                indv.append(random.uniform(bounds[key][1], bounds[key][2]))
           # elif bounds[key][0] == 'grid':
           #     indv.append(random.sample(bounds[key][1],1)[0])
            else:
                raise Exception ('unknown data type is given, either int, float, or grid are allowed for parameter bounds')   
        return np.array(indv)
   
    def eval_bats(self, position_array):
        #---------------------
        # Fitness calcs
        #---------------------
        core_lst=[]
        for case in range (0, position_array.shape[0]):
            core_lst.append(position_array[case, :])
    
        if self.ncores > 1:

            with joblib.Parallel(n_jobs=self.ncores) as parallel:
                fitness_lst=parallel(joblib.delayed(self.fit_worker)(item) for item in core_lst)
                
        else:
            fitness_lst=[]
            for item in core_lst:
                fitness_lst.append(self.fit_worker(item))
        
        return fitness_lst

    def select(self, pos, fit):
        #this function selects the best fitness and position in a population
        best_fit=np.min(fit)
        min_idx=np.argmin(fit)
        best_pos=pos[min_idx,:]
        
        return best_pos, best_fit 
    
    def ensure_bounds(self, vec, bounds):
    
        vec_new = []
        # cycle through each variable in vector 
        for i, (key, val) in enumerate(bounds.items()):
    
            # variable exceedes the minimum boundary
            if vec[i] < bounds[key][1]:
                vec_new.append(bounds[key][1])
    
            # variable exceedes the maximum boundary
            if vec[i] > bounds[key][2]:
                vec_new.append(bounds[key][2])
    
            # the variable is fine
            if bounds[key][1] <= vec[i] <= bounds[key][2]:
                vec_new.append(vec[i])
                
        return vec_new
    
    def fit_worker(self, x):
        #This worker is for parallel calculations
        
        # Clip the bat with position outside the lower/upper bounds and return same position
        x=self.ensure_bounds(x,self.bounds)
        
        if self.grid_flag:
            #decode the individual back to the int/float/grid mixed space
            x=decode_discrete_to_grid(x,self.orig_bounds,self.bounds_map)
        
        # Calculate objective function for each search agent
        fitness = self.fit(x)
        
        return fitness
    
    def ensure_discrete(self, vec):
        #"""
        #to mutate a vector if discrete variables exist 
        #handy function to be used three times within BAT phases

        #Params:
        #vec - bat position in vector/list form

        #Return:
        #vec - updated bat position vector with discrete values
        #"""
        
        for dim in range(self.dim):
            if self.var_type[dim] == 'int':
                vec[dim] = mutate_discrete(x_ij=vec[dim], 
                                               x_min=min(vec),
                                               x_max=max(vec),
                                               lb=self.lb[dim], 
                                               ub=self.ub[dim],
                                               alpha=self.a,
                                               method=self.int_transform,
                                               )
        
        return vec
    
    def Levy(self, dim):
        #function to return levy step
        beta = 1.5
        sigma = (
            math.gamma(1 + beta)
            * math.sin(math.pi * beta / 2)
            / (math.gamma((1 + beta) / 2) * beta * 2 ** ((beta - 1) / 2))
        ) ** (1 / beta)
        u = 0.01 * np.random.randn(dim) * sigma
        v = np.random.randn(dim)
        zz = np.power(np.absolute(v), (1 / beta))
        step = np.divide(u, zz)
        return step

    def evolute(self, ngen, x0=None, verbose=False):
        """
        This function evolutes the BAT algorithm for number of generations.
        
        :param ngen: (int) number of generations to evolute
        :param x0: (list of lists) initial position of the bats (must be of same size as ``nbats``)
        :param verbose: (bool) print statistics to screen
		
        :return: (tuple) (best individual, best fitness, and dictionary containing major search results)
        """
        self.history = {'local_fitness':[], 'global_fitness':[], 'A': [], 'r': []}
        self.fbest=float("inf")
        self.verbose=verbose
        self.Positions = np.zeros((self.nbats, self.dim))
        self.r=self.r0
        if x0:
            assert len(x0) == self.nbats, '--error: the length of x0 ({}) MUST equal the number of bats in the group ({})'.format(len(x0), self.nbats)
            for i in range(self.nbats):
                check_mixed_individual(x=x0[i], bounds=self.orig_bounds) #assert the type provided is consistent
                if self.grid_flag:
                    self.Positions[i,:] = encode_grid_indv_to_discrete(x0[i], bounds=self.orig_bounds, bounds_map=self.bounds_map)
                else:
                    self.Positions[i,:] = x0[i]
        else:
            # Initialize the positions of bats
            for i in range(self.nbats):
                self.Positions[i,:]=self.init_sample(self.bounds)
        
        #Initialize and evaluate the first bat population
        fitness0=self.eval_bats(position_array=self.Positions)
        x0, f0=self.select(pos=self.Positions,fit=fitness0)
        self.xbest=np.copy(x0)
        
        # Main BAT loop
        for l in range(0, ngen):
            self.a= 1 - l * ((1) / ngen)  #mir: a decreases linearly between 1 to 0, for discrete mutation
            #------------------------------------------------------
            # Stage 1A: Loop over all bats to update the positions
            #------------------------------------------------------
            for i in range(0, self.nbats):
                
                if self.levy_flight:
                    #Eq.(11) make a levy flight jump to increase population diversity
                    self.Positions[i,:]=self.Positions[i,:]+np.multiply(np.random.randn(self.dim), self.Levy(self.dim))
                
                #Eq.(8)-(10)
                f1=((self.fmin-self.fmax)*l/ngen+self.fmax)*random.random()
                f2=((self.fmax-self.fmin)*l/ngen+self.fmin)*random.random()
                betas=random.sample(list(range(0,self.nbats)),4)
                self.Positions[i, :]=self.xbest+f1*(self.Positions[betas[0],:]-self.Positions[betas[1],:])
                +f2*(self.Positions[betas[2],:]-self.Positions[betas[3],:])
                self.Positions[i, :] = self.ensure_bounds(self.Positions[i , :], self.bounds)
                self.Positions[i, :] = self.ensure_discrete(self.Positions[i , :])
            #-----------------------
            #Stage 1B: evaluation
            #-----------------------
            fitness1=self.eval_bats(position_array=self.Positions)
            x1, f1=self.select(pos=self.Positions,fit=fitness1)
            if f1 <= self.fbest:
                self.fbest=f1
                self.xbest=x1.copy()
            #---------------------------------
            #Stage 2A: Generate new positions
            #---------------------------------
            for i in range(0, self.nbats):
                # Pulse rate
                if random.random() > self.r:
                    self.Positions[i, :] = self.xbest + self.A * np.random.uniform(-1,1,self.dim)
                self.Positions[i, :] = self.ensure_bounds(self.Positions[i , :], self.bounds)
                self.Positions[i, :] = self.ensure_discrete(self.Positions[i , :])
            #-----------------------
            #Stage 2B: evaluation
            #-----------------------
            fitness2=self.eval_bats(position_array=self.Positions)
            x2, f2=self.select(pos=self.Positions,fit=fitness2)
            if f2 <= self.fbest:
                self.fbest=f2
                self.xbest=x2.copy()
            #---------------------------------
            #Stage 3A: Generate new positions
            #---------------------------------
            for i in range(0, self.nbats):
                # loudness effect
                if random.random() < self.A:
                    self.Positions[i, :] = self.xbest + self.r * np.random.uniform(-1,1,self.dim)
                self.Positions[i, :] = self.ensure_bounds(self.Positions[i , :], self.bounds)
                self.Positions[i, :] = self.ensure_discrete(self.Positions[i , :])
            #-----------------------
            #Stage 3B: evaluation
            #-----------------------
            fitness3=self.eval_bats(position_array=self.Positions)
            x3, f3=self.select(pos=self.Positions,fit=fitness3)
            if f3 <= self.fbest:
                self.fbest=f3
                self.xbest=x3.copy()
            #---------------------------------
            #Stage 4: Check and update A/r
            #---------------------------------                
            if min(f1, f2, f3) <= self.fbest:
                #update A
                self.A = self.alpha*self.A
                #update r
                self.r = self.r0*(1-math.exp(-self.gamma*l))  
            #---------------------------------
            #Stage 5: post-processing
            #---------------------------------             
            #--mir
            if self.mode=='max':
                self.fitness_best_correct=-self.fbest
                self.local_fitness = -min(f1 , f2 , f3)
            else:
                self.fitness_best_correct=self.fbest
                self.local_fitness = min(f1 , f2 , f3)
            
            self.last_pop=self.Positions.copy()
            self.last_fit=np.array(fitness3).copy()
            self.best_position=self.xbest.copy()
            self.history['local_fitness'].append(self.local_fitness)
            self.history['global_fitness'].append(self.fitness_best_correct)
            self.history['A'].append(self.A)
            self.history['r'].append(self.r)   
            
            # Print statistics
            if self.verbose and i % self.nbats:
                print('^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^')
                print('BAT step {}/{}, nbats={}, Ncores={}'.format((l+1)*self.nbats, ngen*self.nbats, self.nbats, self.ncores))
                print('^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^')
                print('Best Bat Fitness:', np.round(self.fitness_best_correct,6))
                if self.grid_flag:
                    self.bat_decoded = decode_discrete_to_grid(self.best_position, self.orig_bounds, self.bounds_map)
                    print('Best Bat Position:', self.bat_decoded)
                else:
                    print('Best Bat Position:', self.best_position)
                print('Loudness A:', self.A)
                print('Pulse rate r:', self.r)
                print('^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^')
        
        #mir-grid
        if self.grid_flag:
            self.bat_correct = decode_discrete_to_grid(self.best_position, self.orig_bounds, self.bounds_map)
        else:
            self.bat_correct = self.best_position
                
        if self.verbose:
            print('------------------------ BAT Summary --------------------------')
            print('Best fitness (y) found:', self.fitness_best_correct)
            print('Best individual (x) found:', self.bat_correct)
            print('--------------------------------------------------------------')  

        if self.mode=='max':
            self.last_fit=-self.last_fit
        
        #--mir return the last population for restart calculations
        if self.grid_flag:
            self.history['last_pop'] = get_population(self.last_pop, fits=self.last_fit, grid_flag=True, 
                                                     bounds=self.orig_bounds, bounds_map=self.bounds_map)
        else:
            self.history['last_pop'] = get_population(self.last_pop, fits=self.last_fit, grid_flag=False)
            
        return self.bat_correct, self.fitness_best_correct, self.history


"""
Helpers for scripts like run_atari.py.
"""

import os
import warnings

import gym

from neorl.rl.baselines.shared import logger
from neorl.rl.baselines.shared.monitor import Monitor
from neorl.rl.baselines.shared.misc_util import set_global_seeds
from neorl.rl.baselines.shared.atari_wrappers import make_atari, wrap_deepmind
from neorl.rl.baselines.shared.misc_util import mpi_rank_or_zero
from neorl.rl.baselines.shared.vec_env import DummyVecEnv, SubprocVecEnv


def make_vec_env(env_id, n_envs=1, seed=None, start_index=0,
                 monitor_dir=None, wrapper_class=None,
                 env_kwargs=None, vec_env_cls=None, vec_env_kwargs=None):
    """
    Create a wrapped, monitored `VecEnv`.
    By default it uses a `DummyVecEnv` which is usually faster
    than a `SubprocVecEnv`.

    :param env_id: (str or Type[gym.Env]) the environment ID or the environment class
    :param n_envs: (int) the number of environments you wish to have in parallel
    :param seed: (int) the initial seed for the random number generator
    :param start_index: (int) start rank index
    :param monitor_dir: (str) Path to a folder where the monitor files will be saved.
        If None, no file will be written, however, the env will still be wrapped
        in a Monitor wrapper to provide additional information about training.
    :param wrapper_class: (gym.Wrapper or callable) Additional wrapper to use on the environment.
        This can also be a function with single argument that wraps the environment in many things.
    :param env_kwargs: (dict) Optional keyword argument to pass to the env constructor
    :param vec_env_cls: (Type[VecEnv]) A custom `VecEnv` class constructor. Default: None.
    :param vec_env_kwargs: (dict) Keyword arguments to pass to the `VecEnv` class constructor.
    :return: (VecEnv) The wrapped environment
    """
    env_kwargs = {} if env_kwargs is None else env_kwargs
    vec_env_kwargs = {} if vec_env_kwargs is None else vec_env_kwargs

    def make_env(rank):
        def _init():
            if isinstance(env_id, str):
                env = gym.make(env_id)
                if len(env_kwargs) > 0:
                    warnings.warn("No environment class was passed (only an env ID) so `env_kwargs` will be ignored")
            else:
                env = env_id(**env_kwargs)
            if seed is not None:
                env.seed(seed + rank)
                env.action_space.seed(seed + rank)
            # Wrap the env in a Monitor wrapper
            # to have additional training information
            monitor_path = os.path.join(monitor_dir, str(rank)) if monitor_dir is not None else None
            # Create the monitor folder if needed
            if monitor_path is not None:
                os.makedirs(monitor_dir, exist_ok=True)
            env = Monitor(env, filename=monitor_path)
            # Optionally, wrap the environment with the provided wrapper
            if wrapper_class is not None:
                env = wrapper_class(env)
            return env
        return _init

    # No custom VecEnv is passed
    if vec_env_cls is None:
        # Default: use a DummyVecEnv
        vec_env_cls = DummyVecEnv

    return vec_env_cls([make_env(i + start_index) for i in range(n_envs)], **vec_env_kwargs)


def make_atari_env(env_id, num_env, seed, wrapper_kwargs=None,
                   start_index=0, allow_early_resets=True,
                   start_method=None, use_subprocess=False):
    """
    Create a wrapped, monitored VecEnv for Atari.

    :param env_id: (str) the environment ID
    :param num_env: (int) the number of environment you wish to have in subprocesses
    :param seed: (int) the initial seed for RNG
    :param wrapper_kwargs: (dict) the parameters for wrap_deepmind function
    :param start_index: (int) start rank index
    :param allow_early_resets: (bool) allows early reset of the environment
    :param start_method: (str) method used to start the subprocesses.
        See SubprocVecEnv doc for more information
    :param use_subprocess: (bool) Whether to use `SubprocVecEnv` or `DummyVecEnv` when
        `num_env` > 1, `DummyVecEnv` is usually faster. Default: False
    :return: (VecEnv) The atari environment
    """
    if wrapper_kwargs is None:
        wrapper_kwargs = {}

    def make_env(rank):
        def _thunk():
            env = make_atari(env_id)
            env.seed(seed + rank)
            env = Monitor(env, logger.get_dir() and os.path.join(logger.get_dir(), str(rank)),
                          allow_early_resets=allow_early_resets)
            return wrap_deepmind(env, **wrapper_kwargs)
        return _thunk
    set_global_seeds(seed)

    # When using one environment, no need to start subprocesses
    if num_env == 1 or not use_subprocess:
        return DummyVecEnv([make_env(i + start_index) for i in range(num_env)])

    return SubprocVecEnv([make_env(i + start_index) for i in range(num_env)],
                         start_method=start_method)


def make_mujoco_env(env_id, seed, allow_early_resets=True):
    """
    Create a wrapped, monitored gym.Env for MuJoCo.

    :param env_id: (str) the environment ID
    :param seed: (int) the initial seed for RNG
    :param allow_early_resets: (bool) allows early reset of the environment
    :return: (Gym Environment) The mujoco environment
    """
    set_global_seeds(seed + 10000 * mpi_rank_or_zero())
    env = gym.make(env_id)
    env = Monitor(env, os.path.join(logger.get_dir(), '0'), allow_early_resets=allow_early_resets)
    env.seed(seed)
    return env


def make_robotics_env(env_id, seed, rank=0, allow_early_resets=True):
    """
    Create a wrapped, monitored gym.Env for MuJoCo.

    :param env_id: (str) the environment ID
    :param seed: (int) the initial seed for RNG
    :param rank: (int) the rank of the environment (for logging)
    :param allow_early_resets: (bool) allows early reset of the environment
    :return: (Gym Environment) The robotic environment
    """
    set_global_seeds(seed)
    env = gym.make(env_id)
    keys = ['observation', 'desired_goal']
    # TODO: remove try-except once most users are running modern Gym
    try:  # for modern Gym (>=0.15.4)
        from gym.wrappers import FilterObservation, FlattenObservation
        env = FlattenObservation(FilterObservation(env, keys))
    except ImportError:  # for older gym (<=0.15.3)
        from gym.wrappers import FlattenDictWrapper  # pytype:disable=import-error
        env = FlattenDictWrapper(env, keys)
    env = Monitor(
        env, logger.get_dir() and os.path.join(logger.get_dir(), str(rank)),
        info_keywords=('is_success',), allow_early_resets=allow_early_resets)
    env.seed(seed)
    return env


def arg_parser():
    """
    Create an empty argparse.ArgumentParser.

    :return: (ArgumentParser)
    """
    import argparse
    return argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)


def atari_arg_parser():
    """
    Create an argparse.ArgumentParser for run_atari.py.

    :return: (ArgumentParser) parser {'--env': 'BreakoutNoFrameskip-v4', '--seed': 0, '--num-timesteps': int(1e7)}
    """
    parser = arg_parser()
    parser.add_argument('--env', help='environment ID', default='BreakoutNoFrameskip-v4')
    parser.add_argument('--seed', help='RNG seed', type=int, default=0)
    parser.add_argument('--num-timesteps', type=int, default=int(1e7))
    return parser


def mujoco_arg_parser():
    """
    Create an argparse.ArgumentParser for run_mujoco.py.

    :return:  (ArgumentParser) parser {'--env': 'Reacher-v2', '--seed': 0, '--num-timesteps': int(1e6), '--play': False}
    """
    parser = arg_parser()
    parser.add_argument('--env', help='environment ID', type=str, default='Reacher-v2')
    parser.add_argument('--seed', help='RNG seed', type=int, default=0)
    parser.add_argument('--num-timesteps', type=int, default=int(1e6))
    parser.add_argument('--play', default=False, action='store_true')
    return parser


def robotics_arg_parser():
    """
    Create an argparse.ArgumentParser for run_mujoco.py.

    :return: (ArgumentParser) parser {'--env': 'FetchReach-v0', '--seed': 0, '--num-timesteps': int(1e6)}
    """
    parser = arg_parser()
    parser.add_argument('--env', help='environment ID', type=str, default='FetchReach-v0')
    parser.add_argument('--seed', help='RNG seed', type=int, default=0)
    parser.add_argument('--num-timesteps', type=int, default=int(1e6))
    return parser



from neorl import RNEAT
import numpy as np 

def test_rneat():
    
    def Sphere(individual):
        """Sphere test objective function.
                F(x) = sum_{i=1}^d xi^2
                d=1,2,3,...
                Range: [-100,100]
                Minima: 0
        """
        return sum(x**2 for x in individual)
    
    nx=5
    lb=-100
    ub=100
    bounds={}
    for i in range(1,nx+1):
            bounds['x'+str(i)]=['float', -100, 100]
    
    # modify your own NEAT config
    config = {
        'pop_size': 50,
        'num_hidden': 1,
        'activation_mutate_rate': 0.1,
        'survival_threshold': 0.3,
        }
    
    # model config
    rneat=RNEAT(fit=Sphere, bounds=bounds, mode='min', config= config, ncores=1, seed=1)
    #A random initial guess (provide one individual)
    x0 = np.random.uniform(lb,ub,nx)
    x_best, y_best, rneat_hist=rneat.evolute(ngen=5, x0=x0, 
                                             verbose=True, checkpoint_itv=None, 
                                             startpoint=None)
    
test_rneat()


#    This file is part of NEORL.

#    Copyright (c) 2021 Exelon Corporation and MIT Nuclear Science and Engineering
#    NEORL is free software: you can redistribute it and/or modify
#    it under the terms of the MIT LICENSE

#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#    SOFTWARE.

# -*- coding: utf-8 -*-
#"""
#@author: Xubo
#@email: guxubo@alumni.sjtu.edu
#"""
import random
import numpy as np
import joblib
from neorl.evolu.discrete import mutate_discrete, encode_grid_to_discrete 
from neorl.evolu.discrete import decode_discrete_to_grid, encode_grid_indv_to_discrete
from neorl.utils.seeding import set_neorl_seed
from neorl.utils.tools import get_population, check_mixed_individual

class JAYA:
    """
    JAYA algorithm
    
    :param mode: (str) problem type, either "min" for minimization problem or "max" for maximization
    :param bounds: (dict) input parameter type and lower/upper bounds in dictionary form. Example: ``bounds={'x1': ['int', 1, 4], 'x2': ['float', 0.1, 0.8], 'x3': ['float', 2.2, 6.2]}``
    :param fit: (function) the fitness function 
    :param npop: (int) number of individuals in the population
    :param int_transform: (str): method of handling int/discrete variables, choose from: ``nearest_int``, ``sigmoid``, ``minmax``.
    :param ncores: (int) number of parallel processors
    :param seed: (int) random seed for sampling
    """
    
    def __init__(self, mode, bounds, fit, npop=50, int_transform ='nearest_int', ncores=1, seed=None):

        self.seed=seed
        set_neorl_seed(self.seed)

        assert npop > 3, '--eror: size of npop must be more than 3'
        self.npop= npop
        self.bounds=bounds
        self.ncores=ncores

        self.int_transform=int_transform
        if int_transform not in ["nearest_int", "sigmoid", "minmax"]:
            raise ValueError('--error: int_transform entered by user is invalid, must be `nearest_int`, `sigmoid`, or `minmax`')


        self.mode=mode
        if mode == 'max':
            self.fit=fit
        elif mode == 'min':
            def fitness_wrapper(*args, **kwargs):
                return -fit(*args, **kwargs)
            self.fit = fitness_wrapper
        else:
            raise ValueError('--error: The mode entered by user is invalid, use either `min` or `max`')

        #infer variable types 
        self.var_type = np.array([bounds[item][0] for item in bounds])
        
        #mir-grid
        if "grid" in self.var_type:
            self.grid_flag=True
            self.orig_bounds=bounds  #keep original bounds for decoding
            #print('--debug: grid parameter type is found in the space')
            self.bounds, self.bounds_map=encode_grid_to_discrete(self.bounds) #encoding grid to int
            #define var_types again by converting grid to int
            self.var_type = np.array([self.bounds[item][0] for item in self.bounds])
        else:
            self.grid_flag=False
            self.bounds = bounds
            self.orig_bounds=bounds
        
        self.dim = len(bounds)
        self.lb=[self.bounds[item][1] for item in self.bounds]
        self.ub=[self.bounds[item][2] for item in self.bounds]
        
    def gen_indv(self, bounds): # individual 

        indv = []
        for key in bounds:
            if bounds[key][0] == 'int':
                indv.append(random.randint(bounds[key][1], bounds[key][2]))
            elif bounds[key][0] == 'float':
                indv.append(random.uniform(bounds[key][1], bounds[key][2]))
            elif bounds[key][0] == 'grid':
                indv.append(random.sample(bounds[key][1],1)[0])
        return indv

    def init_population(self, x0=None, verbose=False): # population

        pop = []
        if x0: # have premary solution
            if verbose:
                print('The first individual provided by the user:', x0[0])
                print('The last individual provided by the user:', x0[-1])
            for i in range(self.npop):
                check_mixed_individual(x=x0[i], bounds=self.orig_bounds) #assert the type provided is consistent
                if self.grid_flag:
                    pop.append(encode_grid_indv_to_discrete(x0[i], bounds=self.orig_bounds, bounds_map=self.bounds_map))
                else:
                    pop.append(x0[i])
        else: # random init
            for i in range(self.npop):
                indv=self.gen_indv(self.bounds)
                pop.append(indv)
        # array
        pop = np.array(pop)

        return pop

    def ensure_bounds(self, vec): # bounds check

        vec_new = []

        for i, (key, val) in enumerate(self.bounds.items()):
            # less than minimum 
            if vec[i] < self.bounds[key][1]:
                vec_new.append(self.bounds[key][1])
            # more than maximum
            if vec[i] > self.bounds[key][2]:
                vec_new.append(self.bounds[key][2])
            # fine
            if self.bounds[key][1] <= vec[i] <= self.bounds[key][2]:
                vec_new.append(vec[i])
        
        return vec_new
    
    def eval_pop(self, pos_array):
        #"""
        #Evaluate fitness of the population with parallel processing.

        #Return:
        #list - pop fitnesses
        #"""
        if self.ncores > 1:
            with joblib.Parallel(n_jobs=self.ncores) as parallel:
                fitness_lst = parallel(joblib.delayed(self.fit_worker)(pos_array[i, :]) for i in range(self.npop))
        else:
            fitness_lst = []
            for i in range(self.npop):
                fitness_lst.append(self.fit_worker(pos_array[i, :]))
        return fitness_lst

    def fit_worker(self, x):
        
        # Clip the wolf with position outside the lower/upper bounds and return same position
        # x=self.ensure_bounds(x)
        
        if self.grid_flag:
            #decode the individual back to the int/float/grid mixed space
            x=decode_discrete_to_grid(x,self.orig_bounds,self.bounds_map)

        fitness = self.fit(x)

        return fitness
    
    def ensure_discrete(self, vec):
        #"""
        #to mutate a vector if discrete variables exist 

        #Params:
        #vec - position in vector/list form

        #Return:
        #vec - updated position vector with discrete values
        #"""
        
        for dim in range(self.dim):
            if self.var_type[dim] == 'int':
                vec[dim] = mutate_discrete(x_ij=vec[dim], 
                                               x_min=min(vec),
                                               x_max=max(vec),
                                               lb=self.lb[dim], 
                                               ub=self.ub[dim],
                                               alpha=self.b,
                                               method=self.int_transform,
                                               )
        
        return vec

    def evolute(self, ngen, x0=None, verbose=False):
        """
        This function evolutes the MFO algorithm for number of generations.
        
        :param ngen: (int) number of generations to evolute
        :param x0: (list of lists) the initial individuals of the population
        :param verbose: (bool) print statistics to screen
        
        :return: (tuple) (best individual, best fitness, and a list of fitness history)
        """
        self.history = {'local_fitness':[], 'global_fitness':[]}
        N = self.npop # population size
        dim = len(self.bounds) # individual length

        set_neorl_seed(self.seed)
        
        fitness_mat = np.zeros(N)

        Best_pos = np.zeros(dim)
        Best_score = float('-inf') # find a maximum, so the larger the better
        Worst_pos = np.zeros(dim)
        Worst_score = float('inf')
        ## INITIALIZE
        #  population
        if x0:
            assert len(x0) == N, '--error: the length of x0 ({}) (initial population) must equal to number of individuals npop ({})'.format(len(x0), self.npop)
            pos = self.init_population(x0=x0, verbose=verbose)
        else:
            pos = self.init_population(verbose=verbose)

        pos = pos*1.0   #this is to account for mixed intger-cont. problems, data needs to be float
        
        # calulate fitness 
        fitness_mat=self.eval_pop(pos)
        for i in range(N):
            if fitness_mat[i] > Best_score:
                Best_score = fitness_mat[i]
                Best_pos = pos[i, :]
            if fitness_mat[i] < Worst_score:
                Worst_score = fitness_mat[i]
                Worst_pos = pos[i, :]

        ## main loop
        best_scores = []
        for gen in range(1, ngen+1):
            self.b= 1 - gen * ((1) / ngen)  #mir: b decreases linearly between 1 to 0, for discrete mutation
            new_pos = np.zeros((N,dim))

            # update pos
            for i in range(N):
                r1=np.random.random(dim)
                r2=np.random.random(dim)
                # Update pos
                new_pos[i,:] = (
                    pos[i,:] 
                    + r1*(Best_pos - abs(pos[i,:]))
                    - r2*(Worst_pos - abs(pos[i,:])) # !! minus
                )
                # check bounds            
                new_pos[i,:] = self.ensure_bounds(new_pos[i,:])
                new_pos[i,:] = self.ensure_discrete(new_pos[i,:])
            
            
            fitness_new=self.eval_pop(new_pos)
                        
            for i in range(N):
                if fitness_new[i] > fitness_mat[i]:
                    pos[i,:] = new_pos[i,:]
                    fitness_mat[i] = fitness_new[i]

            # update best_score and worst_score
            for i in range(N):
                if fitness_mat[i] > Best_score:
                    Best_score = fitness_mat[i]
                    Best_pos = pos[i, :]
                if fitness_mat[i] < Worst_score:
                    Worst_score = fitness_mat[i]
                    Worst_pos = pos[i, :]            

            #-----------------------------
            #Fitness saving 
            #-----------------------------
            self.last_pop=new_pos.copy()
            self.last_fit=np.array(fitness_new).copy()  
            
            gen_avg = sum(fitness_mat) / N                   # current generation avg. fitness
            y_best = Best_score                                # fitness of best individual
            x_best = Best_pos.copy()
            best_scores.append(y_best)
            
            #--mir  show the value wrt min/max
            if self.mode=='min':
                y_best_correct=-y_best
                gen_avg=-gen_avg
                self.history['local_fitness'].append(-np.max(fitness_new))
            else:
                y_best_correct=y_best
                self.history['local_fitness'].append(np.max(fitness_new))

            if verbose:
                print('************************************************************')
                print('JAYA step {}/{}, Ncores={}'.format(gen*self.npop, ngen*self.npop, self.ncores))
                print('************************************************************')
                print('Best fitness:', np.round(y_best_correct,6))

                if self.grid_flag:
                    x_decoded = decode_discrete_to_grid(x_best, self.orig_bounds, self.bounds_map)
                    print('Best individual:', x_decoded)
                else:
                    print('Best individual:', x_best)
                
                print('Average fitness:', np.round(gen_avg,6))
                print('************************************************************')

        #mir-grid
        if self.grid_flag:
            x_best_correct = decode_discrete_to_grid(x_best, self.orig_bounds, self.bounds_map)
        else:
            x_best_correct = x_best
            
        if verbose:
            print('------------------------ JAYA Summary --------------------------')
            print('Best fitness (y) found:', y_best_correct)
            print('Best individual (x) found:', x_best_correct)
            print('--------------------------------------------------------------')
    
        if self.mode=='min':
            best_scores=[-item for item in best_scores]
            self.last_fit=-self.last_fit
        
        self.history['global_fitness'] = best_scores
        
        #--mir return the last population for restart calculations
        if self.grid_flag:
            self.history['last_pop'] = get_population(self.last_pop, fits=self.last_fit, grid_flag=True, 
                                                     bounds=self.orig_bounds, bounds_map=self.bounds_map)
        else:
            self.history['last_pop'] = get_population(self.last_pop, fits=self.last_fit, grid_flag=False)
        
        return x_best_correct, y_best_correct, self.history



#    This file is part of NEORL.

#    Copyright (c) 2021 Exelon Corporation and MIT Nuclear Science and Engineering
#    NEORL is free software: you can redistribute it and/or modify
#    it under the terms of the MIT LICENSE

#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#    SOFTWARE.


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Feb 26 09:33:45 2020

@author: majdi
"""
import os, random
#import tensorflow as tf
#tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)
# import input parameters from the user 
from neorl.parsers.PARSER import InputChecker
from neorl.rl.runners.dqn import DQNAgent
from neorl.rl.runners.ppo2 import PPOAgent
from neorl.rl.runners.a2c import A2CAgent
from neorl.rl.runners.acer import ACERAgent
from multiprocessing import Process
from neorl.evolu.runners.ga import GAAgent
from neorl.evolu.runners.sa import SAAgent
from neorl.utils.neorlcalls import SavePlotCallback
from neorl.rl.baselines.shared.callbacks import BaseCallback


class MultiProc (InputChecker):
    
     def __init__ (self, inp):
    
         self.inp=inp
         os.environ["KMP_WARNINGS"] = "FALSE"
         #os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3' 
     
     def dqn_proc(self):
         dqn_callback=SavePlotCallback(check_freq=self.inp.dqn_dict["check_freq"][0], avg_step=self.inp.dqn_dict["avg_episodes"][0], 
                                       log_dir=self.inp.gen_dict["log_dir"]+self.inp.dqn_dict["casename"][0], plot_mode=self.inp.gen_dict["plot_mode"][0],
                                       total_timesteps=self.inp.dqn_dict["time_steps"][0], basecall=BaseCallback())
         dqn=DQNAgent(self.inp, dqn_callback)
         dqn.build()
         
         return

     def ppo_proc(self):
         ppo_callback=SavePlotCallback(check_freq=self.inp.ppo_dict["check_freq"][0], avg_step=self.inp.ppo_dict["avg_episodes"][0], 
                                       log_dir=self.inp.gen_dict["log_dir"]+self.inp.ppo_dict["casename"][0], plot_mode=self.inp.gen_dict["plot_mode"][0],
                                       total_timesteps=self.inp.ppo_dict["time_steps"][0], basecall=BaseCallback())
         ppo=PPOAgent(self.inp, ppo_callback)
         ppo.build()
         
         return

     def a2c_proc(self):
         a2c_callback=SavePlotCallback(check_freq=self.inp.a2c_dict["check_freq"][0], avg_step=self.inp.a2c_dict["avg_episodes"][0], 
                                       log_dir=self.inp.gen_dict["log_dir"]+self.inp.a2c_dict["casename"][0], plot_mode=self.inp.gen_dict["plot_mode"][0],
                                       total_timesteps=self.inp.a2c_dict["time_steps"][0], basecall=BaseCallback())
         a2c=A2CAgent(self.inp, a2c_callback)
         a2c.build()
         
         return

     def acer_proc(self):
         acer_callback=SavePlotCallback(check_freq=self.inp.acer_dict["check_freq"][0], avg_step=self.inp.acer_dict["avg_episodes"][0], 
                                       log_dir=self.inp.gen_dict["log_dir"]+self.inp.acer_dict["casename"][0], plot_mode=self.inp.gen_dict["plot_mode"][0],
                                       total_timesteps=self.inp.acer_dict["time_steps"][0], basecall=BaseCallback())
         acer=ACERAgent(self.inp, acer_callback)
         acer.build()
         
         return

     def ga_proc(self):
         #check_freq is set by default to every generation for GA, so it does not have any effect in the callback 
         #total_timesteps is set by default to all generations for GA,  so it does not have any effect in the callback 
         ga_callback=SavePlotCallback(check_freq=self.inp.ga_dict["check_freq"][0], avg_step=self.inp.ga_dict["pop"][0], 
                               log_dir=self.inp.gen_dict["log_dir"]+self.inp.ga_dict["casename"][0], plot_mode=self.inp.gen_dict["plot_mode"][0],
                               total_timesteps=self.inp.ga_dict["ngen"][0], basecall=BaseCallback())
         ga=GAAgent(self.inp, ga_callback)
         ga.build()
         
         return
     
     def sa_proc(self):
         #avg_step is set by default to every check_freq, so it does not have any effect in the callback 
         #total_timesteps is set by default to all generations for SA,  so it does not have any effect in the callback 
         sa_callback=SavePlotCallback(check_freq=self.inp.sa_dict["check_freq"][0], avg_step=self.inp.sa_dict["avg_step"][0], 
                               log_dir=self.inp.gen_dict["log_dir"]+self.inp.sa_dict["casename"][0], plot_mode=self.inp.gen_dict["plot_mode"][0],
                               total_timesteps=self.inp.sa_dict["steps"][0], basecall=BaseCallback())
         sa=SAAgent(self.inp, sa_callback)
         sa.build()
         
         return

     def run_all(self):
        
        # setup all processes
        if self.inp.dqn_dict['flag'][0]:
            dqn_task = Process(name='dqn', target=self.dqn_proc)
            
        if self.inp.ppo_dict['flag'][0]:
            ppo_task = Process(name='ppo', target=self.ppo_proc)

        if self.inp.a2c_dict['flag'][0]:
            a2c_task = Process(name='a2c', target=self.a2c_proc)       

        if self.inp.acer_dict['flag'][0]:
            acer_task = Process(name='acer', target=self.acer_proc)
            
        if self.inp.ga_dict['flag'][0]:
            ga_task = Process(name='ga', target=self.ga_proc)   

        if self.inp.sa_dict['flag'][0]:
            sa_task = Process(name='sa', target=self.sa_proc)
            
        # start running processes
        if self.inp.dqn_dict['flag'][0]:
            dqn_task.start()
            print('--- DQN is running on {cores} core(s)'.format(cores=self.inp.dqn_dict["ncores"][0]))
            
        if self.inp.ppo_dict['flag'][0]:
            ppo_task.start()
            print('--- PPO is running on {cores} core(s)'.format(cores=self.inp.ppo_dict["ncores"][0]))

        if self.inp.a2c_dict['flag'][0]:
            a2c_task.start()
            print('--- A2C is running on {cores} core(s)'.format(cores=self.inp.a2c_dict["ncores"][0]))

        if self.inp.acer_dict['flag'][0]:
            acer_task.start()
            print('--- ACER is running on {cores} core(s)'.format(cores=self.inp.acer_dict["ncores"][0]))
            
        if self.inp.ga_dict['flag'][0]:
            ga_task.start()
            print('--- GA is running on {cores} core(s)'.format(cores=self.inp.ga_dict["ncores"][0]))
            
        if self.inp.sa_dict['flag'][0]:
            sa_task.start()
            print('--- SA is running on {cores} core(s)'.format(cores=self.inp.sa_dict["ncores"][0]))
        
        print('------------------------------------------------------------------------------')
    
    



from neorl import CS

def test_cs():
    #Define the fitness function
    def Sphere(individual):
        """Sphere test objective function.
                        F(x) = sum_{i=1}^d xi^2
                        d=1,2,3,...
                        Range: [-100,100]
                        Minima: 0
        """
        y=sum(x**2 for x in individual)
        return y
    #Setup the parameter space (d=5)
    nx=5
    BOUNDS={}
    for i in range(1,nx+1):
        BOUNDS['x'+str(i)]=['float', -100, 100]
    
    #setup and evolute CS
    cs = CS(mode = 'min', bounds = BOUNDS, fit = Sphere, ncuckoos = 40, pa = 0.25, seed=1)
    x_best, y_best, cs_hist=cs.evolute(ngen = 200, verbose=False)

test_cs()


#!/usr/bin/env python3
import warnings

from neorl.rl.baselines.shared import logger
from neorl.rl.baselines.acer.acer import ACER
from neorl.rl.baselines.shared.policies import CnnPolicy, CnnLstmPolicy
from neorl.rl.baselines.shared.cmd_util import make_atari_env, atari_arg_parser
from neorl.rl.baselines.shared.vec_env import VecFrameStack


def train(env_id, num_timesteps, seed, policy, lr_schedule, num_cpu):
    """
    train an ACER model on atari

    :param env_id: (str) Environment ID
    :param num_timesteps: (int) The total number of samples
    :param seed: (int) The initial seed for training
    :param policy: (A2CPolicy) The policy model to use (MLP, CNN, LSTM, ...)
    :param lr_schedule: (str) The type of scheduler for the learning rate update ('linear', 'constant',
                                 'double_linear_con', 'middle_drop' or 'double_middle_drop')
    :param num_cpu: (int) The number of cpu to train on
    """
    env = VecFrameStack(make_atari_env(env_id, num_cpu, seed), 4)
    if policy == 'cnn':
        policy_fn = CnnPolicy
    elif policy == 'lstm':
        policy_fn = CnnLstmPolicy
    else:
        warnings.warn("Policy {} not implemented".format(policy))
        return

    model = ACER(policy_fn, env, lr_schedule=lr_schedule, buffer_size=5000, seed=seed)
    model.learn(total_timesteps=int(num_timesteps * 1.1))
    env.close()
    # Free memory
    del model


def main():
    """
    Runs the test
    """
    parser = atari_arg_parser()
    parser.add_argument('--policy', choices=['cnn', 'lstm', 'lnlstm'], default='cnn', help='Policy architecture')
    parser.add_argument('--lr_schedule', choices=['constant', 'linear'], default='constant',
                        help='Learning rate schedule')
    parser.add_argument('--logdir', help='Directory for logging')
    args = parser.parse_args()
    logger.configure(args.logdir)
    train(args.env, num_timesteps=args.num_timesteps, seed=args.seed,
          policy=args.policy, lr_schedule=args.lr_schedule, num_cpu=16)


if __name__ == '__main__':
    main()



from neorl.rl.baselines.shared import logger 
from neorl.rl.baselines.acktr import ACKTR
from neorl.rl.baselines.shared.cmd_util import make_atari_env, atari_arg_parser
from neorl.rl.baselines.shared.vec_env.vec_frame_stack import VecFrameStack
from neorl.rl.baselines.shared.policies import CnnPolicy


def train(env_id, num_timesteps, seed, num_cpu):
    """
    train an ACKTR model on atari

    :param env_id: (str) Environment ID
    :param num_timesteps: (int) The total number of samples
    :param seed: (int) The initial seed for training
    :param num_cpu: (int) The number of cpu to train on
    """
    env = VecFrameStack(make_atari_env(env_id, num_cpu, seed), 4)
    model = ACKTR(CnnPolicy, env, nprocs=num_cpu, seed=seed)
    model.learn(total_timesteps=int(num_timesteps * 1.1))
    env.close()


def main():
    """
    Runs the test
    """
    args = atari_arg_parser().parse_args()
    logger.configure()
    train(args.env, num_timesteps=args.num_timesteps, seed=args.seed, num_cpu=32)


if __name__ == '__main__':
    main()



#    This file is part of NEORL.

#    Copyright (c) 2021 Exelon Corporation and MIT Nuclear Science and Engineering
#    NEORL is free software: you can redistribute it and/or modify
#    it under the terms of the MIT LICENSE

#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#    SOFTWARE.

# -*- coding: utf-8 -*-
#"""
#Created on Tue Jul 13 21:06:40 2021
#
#@author: Paul Seurin
#"""

########################
#
# Import Packages
#
########################

import gym
from gym.spaces import Discrete, Box, MultiDiscrete
import numpy as np
import copy
import matplotlib.pyplot as plt
import random

##############################
#
# Define Knapsack Function
#        (Discrete)
#
##############################

class KP(gym.Env):

  def __init__ (self,obj_list,weight_capacity,optimum_knapsack = None, episode_length = None, method = None):                
    self.obj_library = {}
    self.weight_capacity = weight_capacity
    self.obj_weight_to_value = {}
    
    for i in range(len(obj_list)):
        self.obj_library["%d"%(i + 1)] = [obj_list[i][0],obj_list[i][1]]
        w_to_v = obj_list[i][1] / obj_list[i][0]
        self.obj_weight_to_value["%d"%(i + 1)] = w_to_v
    
    self.number_of_objects = len(obj_list)
    self.state = - 10**6.0 * np.ones((self.number_of_objects,3), dtype = float) # values/weight/weight to value ratio
    
    self.episode_length = episode_length # Number of kp solved per episodes
    self.obj_id = copy.deepcopy(list(self.obj_library.keys())) # list with the ids of the objects
    self.obj_list = [] # List representing a bag of objects
    self.full_obj_list = copy.deepcopy(np.transpose(obj_list)) # store all the objects for plotting purposes
    self.counter = 0 # counter to record a knap
    self.subcounter = 0 # counter to record each knap per episode
    self.method = method #RL method used for optimization
    if self.method not in ['ppo', 'a2c', 'acktr', 'neat', 'dqn', 'acer']:
        raise ValueError("--error: The only methods available are 'ppo', 'a2c', 'acktr', 'neat', 'dqn', or 'acer' not {}".format(self.method))
    if self.method in ['ppo', 'a2c', 'acktr', 'neat']:
        self.action_bounds = [i for i in range(self.number_of_objects,0,-1)]
        self.action_space = MultiDiscrete(self.action_bounds) # action space 
    elif self.method in ['acer', 'dqn']:
        self.action_bounds = self.number_of_objects
        self.action_space = Discrete(self.action_bounds) # action space 
    self.observation_space = Box(low=0, high=self.number_of_objects, shape=(self.number_of_objects * 3,), dtype=int) # state space  
    self.best_knap = -10**6 # variable to store the value of the best knapsack
    self._iter_episode = 0 # record current episode (for logger)
    self.reset()
    self.done = False
    if optimum_knapsack is not None: # record the (known) best knapsack map and plot it
        optimum_knap_map = - 10**6.0 * np.ones((self.number_of_objects,2), dtype = float) 
        for increment in range(0,len(optimum_knapsack)):
            optimum_knap_map[:,0][increment] = self.obj_library['%d'%(optimum_knapsack[increment])][0]
            optimum_knap_map[:,1][increment] = self.obj_library['%d'%(optimum_knapsack[increment])][1]
        optimum_knap_value = self.Compute_knap_value(knap = optimum_knap_map)
        title_map = "Sum of the values : {} \n".format(optimum_knap_value)
        title_map += "Sum of the weights : {} \n".format(self._check_weight_cap(optimum_knap_map))
        title_map += "Weight's limit : {} \n".format(self.weight_capacity)
        _plot_knap_map(optimum_knap_map,flag = False, name = "Optimum_knap_map_%d.png"%(len(obj_list)), title_map = title_map, all_objects=np.transpose(obj_list))
  def step(self, x): 
    if self.method in ['ppo', 'a2c', 'acktr', 'neat']:
        self.counter = 0 # initialize the per knapsack counter
        for action in x: 
            if self.counter == 0:
                self.reset()
            self.counter += 1
            if action <= len(self.obj_id) - 1:  
                new_obj = self.obj_id[action] # recover the objects corresponding to action 'action'
            else: # ensure a valid action considering that the action space is fixed
                action = random.randint(0,len(self.obj_id) - 1)
                new_obj = self.obj_id[action]

            # place the objects in hte knapsack and fill the state space
            self.state[:,0][self.counter - 1] = self.obj_library[new_obj][0]# --- put as embedding the value
            self.state[:,1][self.counter - 1] = self.obj_library[new_obj][1]# --- put as embedding the weights
            self.state[:,2][self.counter - 1] = self.state[:,1][self.counter - 1] / self.state[:,0][self.counter - 1]
    
            index1 = self.obj_id.index(self.obj_id[action]) # remove the objects chosen by action : 'action' from the action space
            self.obj_id.pop(index1)
            self.obj_list.append(new_obj) # list of objects in the knapsack (for neorl callback)

            if  self._check_weight_cap(self.state) > self.weight_capacity:
                self.state[:,0][self.counter - 1] = - 10**6.0 # --- remove last object
                self.state[:,1][self.counter - 1] = - 10**6.0
                self.state[:,2][self.counter - 1] = - 10**6.0
                break # stop filling the knapsack and evaluate the reward
            elif self.obj_list == []:
                raise ValueError("--error: the weight capacity is higher than the sum of the weights of all the objects. The problem is trivial")

        reward = self._get_stats()
        self.subcounter += 1 
        if self.subcounter != self.episode_length + 1: # generate the new knapsack
            self.obj_id = copy.deepcopy(list(self.obj_library.keys()))
            self.obj_list.pop(-1)
            individual = copy.deepcopy(self.obj_list)
            self.obj_list = []
            self.state[:,0] = - 10**6.0 * np.ones(self.number_of_objects)
            self.state[:,1] = - 10**6.0 * np.ones(self.number_of_objects)
            self.state[:,2] = - 10**6.0 * np.ones(self.number_of_objects)     
    
    elif self.method in ['acer', 'dqn']:
        action = x
        if action <= len(self.obj_id) - 1:  
            new_obj = self.obj_id[action] # recover the object corresponding to action 'action'
        else: # ensure a valid action considering that the action space is fixed
            action = random.randint(0,len(self.obj_id) - 1)
            new_obj = self.obj_id[action]
        self.counter += 1 # --- increment the global counter
        # place the objects in hte knapsack and fill the state space
        self.state[:,0][self.counter - 1] = self.obj_library[new_obj][0]# --- put as embedding the value
        self.state[:,1][self.counter - 1] = self.obj_library[new_obj][1]# --- put as embedding the weights
        self.state[:,2][self.counter - 1] = self.state[:,1][self.counter - 1] / self.state[:,0][self.counter - 1]
    
        index1 = self.obj_id.index(self.obj_id[action]) # remove the objects chosen by action : 'action' from the action space
        self.obj_id.pop(index1)
        self.obj_list.append(new_obj) # list of objects in the knapsack (for neorl callback)
        reward = 0
        if  self._check_weight_cap(self.state) > self.weight_capacity:
            self.state[:,0][self.counter - 1] = - 10**6.0 # --- remove last object
            self.state[:,1][self.counter - 1] = - 10**6.0
            self.state[:,2][self.counter - 1] = - 10**6.0
            reward = self._get_stats()
            self.subcounter += 1
            self.counter = 0
            if self.subcounter != self.episode_length + 1: # generate the new knapsack 
                self.obj_id = copy.deepcopy(list(self.obj_library.keys()))
                self.obj_list.pop(-1)
                individual = copy.deepcopy(self.obj_list)
                self.obj_list = []
                self.state[:,0] = - 10**6.0 * np.ones(self.number_of_objects)
                self.state[:,1] = - 10**6.0 * np.ones(self.number_of_objects)
                self.state[:,2] = - 10**6.0 * np.ones(self.number_of_objects)
        else:
            individual = copy.deepcopy(self.obj_list) # a knapsack is not complete.

    if self.subcounter == self.episode_length:# episode terminates
        self.done = True
        self._iter_episode += 1
        self.subcounter = 0
    return ([self.state.flatten(),reward, self.done, {'x':individual}])

  def reset(self):
    self.done = False
    self.obj_id = copy.deepcopy(list(self.obj_library.keys()))
    self.state[:,0] = - 10**6.0 * np.ones(self.number_of_objects)
    self.state[:,1] = - 10**6.0 * np.ones(self.number_of_objects)
    self.state[:,2] = - 10**6.0 * np.ones(self.number_of_objects)
    return (self.state.flatten())

  def Compute_knap_value(self, knap = None):
    if knap is None:
        knap = copy.deepcopy(self.state)        
    value = 0
    limit_knap = np.where(knap[:,0] == -10**6.0)[0] # the list of objects is either -10**6.0 or a reasonable weights
    if len(limit_knap) > 0:
        limit_knap = limit_knap[0]
    else:
        limit_knap = self.numlocs 
    for increment in range(0,limit_knap): # sum the values of the object chosen
        value += knap[:,0][increment]    
    return value

  def _check_weight_cap(self, knap = None):
    if knap is None:
        knap = copy.deepcopy(self.state) 
    value = 0
    limit_knap = np.where(knap[:,0] == -10**6.0)[0] # the list of objects is either -10**6.0 or a reasonable weights
    if len(limit_knap) > 0:
        limit_knap = limit_knap[0]
    else:
        limit_knap = self.numlocs 
    for increment in range(0,limit_knap): # sum the weights of the object chosen and verify that it does not exceed the weight limit
        value += knap[:,1][increment]    
    return value 

  def _get_stats(self,flag = True):
      score = self.Compute_knap_value() 
      if flag: # compute the best knapsack and plot it only for complete knapsack
        if score - self.best_knap > 0:
            self.best_knap = score
            current_knap_value = self.Compute_knap_value(knap = self.state)
            title_map = " Episode: {}, Value of the Knapsack : {} \n".format(self._iter_episode + 1, np.ceil(current_knap_value))
            title_map += "Sum of the weights : {} \n".format(self._check_weight_cap(self.state))
            _plot_knap_map(self.state,flag = False, name = "Best_Knap_map_%d_%s.png"%(len(np.transpose(self.full_obj_list)),self.method), title_map = title_map, all_objects = self.full_obj_list)

      return score
            
  def render(self, mode = 'human'):
    pass  

def _plot_knap_map(knap_embed,flag = False, name ="Obj_dist.png", title_map = None,all_objects = None):
    """
      Plot Object distribution stored in : knap_embed
    """
    Full_map = copy.deepcopy(knap_embed)
    X_or = Full_map[:,0]
    Y_or = Full_map[:,1]
    limit_knap = np.where(X_or == -10**6.0)[0][0]
    plt.figure()
    if title_map is not None:
        plt.title(title_map)
    else:
        plt.title('Items distribution')
    plt.ylabel('weights [-]')
    plt.xlabel('values [-]')
    plt.scatter(all_objects[0],all_objects[1],color = 'blue', label = "Remaining items")
    plt.scatter(X_or[:limit_knap],Y_or[:limit_knap], color = 'orange' ,label = "Chosen items")
    plt.legend()
    plt.savefig("%s"%(name),format='png' ,dpi=300, bbox_inches="tight")
    plt.close()


#    This file is part of NEORL.

#    Copyright (c) 2021 Exelon Corporation and MIT Nuclear Science and Engineering
#    NEORL is free software: you can redistribute it and/or modify
#    it under the terms of the MIT LICENSE

#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#    SOFTWARE.

# -*- coding: utf-8 -*-
#"""
#Created on Sun Jul 25 20:51:03 2021
#
#@author: Majdi
#"""

#--------------------------------------------------------
# This class is inspired from the work by this repo:
# https://github.com/thieunguyen5991  
#--------------------------------------------------------

from pandas import read_csv
from numpy import sum, max, abs, cos, pi, sqrt, exp, e
from numpy.random import seed, choice, uniform
import os 


class Root:
    def __init__(self, f_name=None, f_shift_data_file=None, f_bias=None):
        self.f_name = f_name
        self.f_shift_data_file = f_shift_data_file
        self.f_bias = f_bias
        self.support_path_data=os.path.join(os.path.dirname(__file__), 'tools/data08.csv')
        
    def load_shift_data(self):
        data = read_csv(self.support_path_data, usecols=[self.f_shift_data_file])
        return data.values.reshape(-1)


class F1(Root):
    def __init__(self, f_name="Shifted Sphere Function", f_shift_data_file="F1", f_bias=-450):
        Root.__init__(self, f_name, f_shift_data_file, f_bias)

    def fit(self, solution=None):
        problem_size = len(solution)
        if problem_size > 1000:
            print("CEC 2008 not support for problem size > 1000")
            return 1
        shift_data = self.load_shift_data()[:problem_size]
        return sum((solution - shift_data)**2) + self.f_bias
    
    def return_global(self, nx):
        
        xmin = self.load_shift_data()[:nx]
        ymin = self.fit(xmin)
        
        return xmin, ymin
        
    
    

class F2(Root):
    def __init__(self, f_name="Schwefel’s Problem 2.21", f_shift_data_file="F2", f_bias=-450):
        Root.__init__(self, f_name, f_shift_data_file, f_bias)

    def fit(self, solution=None):
        problem_size = len(solution)
        if problem_size > 1000:
            print("CEC 2008 not support for problem size > 1000")
            return 1
        shift_data = self.load_shift_data()[:problem_size]

        return max(abs(solution - shift_data)) + self.f_bias

    def return_global(self, nx):
        
        xmin = self.load_shift_data()[:nx]
        ymin = self.fit(xmin)
        
        return xmin, ymin
    
class F3(Root):
    def __init__(self, f_name="Shifted Rosenbrock’s Function", f_shift_data_file="F3", f_bias=390, f_matrix=None):
        Root.__init__(self, f_name, f_shift_data_file, f_bias)
        self.f_matrix = f_matrix

    def fit(self, solution=None):
        problem_size = len(solution)
        if problem_size > 1000:
            print("CEC 2008 not support for problem size > 1000")
            return 1
        shift_data = self.load_shift_data()[:problem_size]
        z = solution - shift_data + 1
        result = 0
        for i in range(0, problem_size-1):
            result += 100*(z[i]**2 - z[i+1])**2 + (z[i] - 1)**2
        return result + self.f_bias

    def return_global(self, nx):
        
        xmin = self.load_shift_data()[:nx]
        ymin = self.fit(xmin)
        
        return xmin, ymin
    
class F4(Root):
    def __init__(self, f_name="Shifted Rastrigin’s Function", f_shift_data_file="F4", f_bias=-330):
        Root.__init__(self, f_name, f_shift_data_file, f_bias)

    def fit(self, solution=None):
        problem_size = len(solution)
        if problem_size > 1000:
            print("CEC 2008 not support for problem size > 1000")
            return 1
        shift_data = self.load_shift_data()[:problem_size]
        z = solution - shift_data
        return sum(z**2 - 10*cos(2*pi*z) + 10) + self.f_bias

    def return_global(self, nx):
        
        xmin = self.load_shift_data()[:nx]
        ymin = self.fit(xmin)
        
        return xmin, ymin   

class F5(Root):
    def __init__(self, f_name="Shifted Griewank’s Function", f_shift_data_file="F5", f_bias=-180):
        Root.__init__(self, f_name, f_shift_data_file, f_bias)

    def fit(self, solution=None):
        problem_size = len(solution)
        if problem_size > 1000:
            print("CEC 2008 not support for problem size > 1000")
            return 1
        shift_data = self.load_shift_data()[:problem_size]
        z = solution - shift_data
        result = sum(z**2/4000)
        temp = 1.0
        for i in range(0, problem_size):
            temp *= cos(z[i] / sqrt(i+1))
        return result - temp + 1 + self.f_bias

    def return_global(self, nx):
        
        xmin = self.load_shift_data()[:nx]
        ymin = self.fit(xmin)
        
        return xmin, ymin
    
class F6(Root):
    def __init__(self, f_name="Shifted Ackley’s Function", f_shift_data_file="F6", f_bias=-140):
        Root.__init__(self, f_name, f_shift_data_file, f_bias)

    def fit(self, solution=None):
        problem_size = len(solution)
        if problem_size > 1000:
            print("CEC 2008 not support for problem size > 1000")
            return 1
        shift_data = self.load_shift_data()[:problem_size]
        z = solution - shift_data
        return -20*exp(-0.2*sqrt(sum(z**2)/problem_size)) - exp(sum(cos(2*pi*z))/problem_size) + 20 + e + self.f_bias
    
    def return_global(self, nx):
        
        xmin = self.load_shift_data()[:nx]
        ymin = self.fit(xmin)
        
        return xmin, ymin
    
class F7(Root):
    def __init__(self, f_name="FastFractal “DoubleDip” Function", f_shift_data_file=None, f_bias=None, f_matrix=None):
        Root.__init__(self, f_name, f_shift_data_file, f_bias)
        self.f_matrix = f_matrix

    def fit(self, solution=None):
        seed(0)
        problem_size = len(solution)
        if problem_size > 1000:
            print("CEC 2008 not support for problem size > 1000")
            return 1

        def __doubledip__(x, c, s):
            if -0.5 < x < 0.5:
                return (-6144*(x - c)**6 + 3088*(x - c)**4 - 392*(x - c)**2 + 1)*s
            else:
                return 0

        def __fractal1d__(x):
            result1 = 0.0
            for k in range(1, 4):
                result2 = 0.0
                upper = 2**(k-1)
                for t in range(1, upper):
                    selected = choice([0, 1, 2])
                    result2 += sum([ __doubledip__(x, uniform(), 1.0 / (2**(k-1)*(2-uniform()))) for _ in range(0, selected)])
                result1 += result2
            return result1

        def __twist__(y):
            return 4*(y**4 - 2*y**3 + y**2)

        result = solution[-1] + __twist__(solution[0])
        for i in range(0, problem_size-1):
            x = solution[i] + __twist__(solution[i%problem_size + 1])
            result += __fractal1d__(x)
        return result


#    This file is part of NEORL.

#    Copyright (c) 2021 Exelon Corporation and MIT Nuclear Science and Engineering
#    NEORL is free software: you can redistribute it and/or modify
#    it under the terms of the MIT LICENSE

#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#    SOFTWARE.

#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#"""
#Created on Tue Feb 25 14:42:24 2020
#
#@author: majdi
#"""

import warnings
warnings.filterwarnings("ignore")
import random
import gym
import pandas as pd
import numpy as np
from neorl.parsers.PARSER import InputChecker
import copy, os

class SAAgent(InputChecker):
    
    def __init__ (self, inp, callback):    
        
        """
        Input: 
        inp: is a dictionary of validated user input {"ncores": 8, "env": 6x6, ...}
        callback: a class of callback built from stable-baselines to allow intervening during training 
                  to process data and save models
        """
        self.inp= inp 
        self.env = gym.make(self.inp.gen_dict['env'][0], casename=self.inp.sa_dict['casename'][0], exepath=self.inp.gen_dict['exepath'][0], 
                            log_dir=self.inp.gen_dict['log_dir'], env_data=self.inp.gen_dict['env_data'][0])
        self.fit=self.env.fit
        
        np.random.seed(10)
        self.log_dir=self.inp.gen_dict['log_dir']+self.inp.sa_dict["casename"][0]
        self.callback=callback
        
        #--------------------------------
        # Infer SA user paramters from the sa_dict
        #--------------------------------
        self.Tmax=self.inp.sa_dict['Tmax'][0]
        self.Tmin=self.inp.sa_dict['Tmin'][0]
        self.steps=self.inp.sa_dict['steps'][0]
        self.swap=self.inp.sa_dict['swap'][0]
        self.check_freq=self.inp.sa_dict['check_freq'][0]
        self.lbound=self.inp.sa_dict['lbound'][0]
        self.ubound=self.inp.sa_dict['ubound'][0]
        self.lbound=[int(item) for item in self.lbound]
        self.ubound=[int(item) for item in self.ubound]
        self.ubound=[item+1 for item in self.ubound]  # adding +1 to the ubound for np.random.randint --> samples between [lbound, ubound)
        self.indpb=self.inp.sa_dict['indpb'][0]
        self.cooling=self.inp.sa_dict['cooling'][0]
        self.x_index=list(range(self.inp.gen_dict['xsize'][0])) # list of all indices of input space
        
        #--------------------------------
        # KBS mode (via a CSV dataset)
        #--------------------------------
        if self.swap=='kbs':
            print('--debug: A path to KBS dataset is provided for SA')
            self.indpb_kbs=self.inp.sa_dict['indpb_kbs'][0]
            self.kbs_path=self.inp.sa_dict['kbs_path'][0]
            if not os.path.exists (self.kbs_path):
                raise Exception ('a kbs mode is used for SA, but the path to dataset ({}) does not exist'.format(self.kbs_path))
            print('--debug: Loading the dataset from {} ...'.format(self.kbs_path))
            self.kbs_data=pd.read_csv(self.kbs_path).values
            self.ind_list=list(range(0,self.kbs_data.shape[0]))
        
        self.kbs_usage=0
        self.real_good=0
        
    def singleswap(self) :
        #np.random.seed(10)
        # this function swaps two indices in the input vector, sutiable for input shuffling
        
        dual_index=random.sample(self.x_index,2)
        index1, index2  = dual_index[0], dual_index[1]
        
        if self.lbound == None and self.ubound == None:   #swap only
            self.x[index1], self.x[index2] = self.x[index2], self.x[index1]
        else: 
            
            if len(self.lbound) == 1 and len(self.ubound) == 1:
                sample=np.random.randint(self.lbound, self.ubound)
            else:
                sample=np.random.randint(self.lbound[index1], self.ubound[index1])
                
            while self.x[index1] == sample:  # Resample if no change exists
                if len(self.lbound) == 1 and len(self.ubound) == 1:
                    sample=np.random.randint(self.lbound, self.ubound)
                else:
                    sample=np.random.randint(self.lbound[index1], self.ubound[index1])
                
            self.x[index1] = sample            
        
        return
     
    def dualswap(self) :
        #np.random.seed(10)
        # this function swaps two indices in the input vector, sutiable for input shuffling
        
        dual_index=random.sample(self.x_index,2)
        index1, index2  = dual_index[0], dual_index[1]
        
        if self.lbound == None and self.ubound == None:
            self.x[index1], self.x[index2] = self.x[index2], self.x[index1]
        else:
            if len(self.lbound) == 1 and len(self.ubound) == 1:
                sample1, sample2 = np.random.randint(self.lbound, self.ubound), np.random.randint(self.lbound, self.ubound)
            else:
                sample1, sample2 = np.random.randint(self.lbound[index1], self.ubound[index1]), np.random.randint(self.lbound[index2], self.ubound[index2])
                
            while self.x[index1] == sample1 or self.x[index2] == sample2:
                if len(self.lbound) == 1 and len(self.ubound) == 1:
                    sample1, sample2 = np.random.randint(self.lbound, self.ubound), np.random.randint(self.lbound, self.ubound)
                else:
                    sample1, sample2 = np.random.randint(self.lbound[index1], self.ubound[index1]), np.random.randint(self.lbound[index2], self.ubound[index2])

            self.x[index1], self.x[index2] = sample1, sample2

        return

    
    def quadswap(self) :
        #np.random.seed(10)
        # this function swaps four indices in the input vector, sutiable for shuffling
        
        quad_index=random.sample(self.x_index,4)
        index1, index2, index3, index4 = quad_index[0], quad_index[1], quad_index[2], quad_index[3] 
        
        if self.lbound == None and self.ubound == None:
            self.x[index1], self.x[index2], self.x[index3], self.x[index4] = self.x[index2], self.x[index1], self.x[index4], self.x[index3]
        elif len(self.lbound) == 1 and len(self.ubound) == 1:
            sample1, sample2 = np.random.randint(self.lbound, self.ubound+1), np.random.randint(self.lbound, self.ubound+1)
            sample3, sample4 = np.random.randint(self.lbound, self.ubound+1), np.random.randint(self.lbound, self.ubound+1)
            while self.x[index1] == sample1 or self.x[index2] == sample2:
                sample1, sample2 = np.random.randint(self.lbound, self.ubound+1), np.random.randint(self.lbound, self.ubound+1)
            while self.x[index3] == sample3 or self.x[index4] == sample4:
                sample3, sample4 = np.random.randint(self.lbound, self.ubound+1), np.random.randint(self.lbound, self.ubound+1)
            self.x[index1], self.x[index2] = sample1, sample2
            self.x[index3], self.x[index4] = sample3, sample4
                                              
        return 

    def fullswap(self) :
        #np.random.seed(10)
        # this function randomly shuffels all data in the input vector
        if self.lbound == None and self.ubound == None:
             self.x=random.sample(self.x,len(self.x))
        elif len(self.lbound) == 1 and len(self.ubound) == 1:
            self.x=np.random.randint(self.lbound, self.ubound,len(self.x))
        else:
            for i in range(len(self.x)):
                if random.random() < self.indpb:
                    self.x[i]=np.random.randint(self.lbound[i], self.ubound[i])
        return
    
    def kbs(self):
        
        try:
            if random.random() < self.indpb_kbs:
                
                ind=random.sample(self.ind_list,1)[0]
                self.ind_list.remove(ind)  # remove the new location from the list
                if not self.ind_list:
                    print("--- KBS Indices are reinitialized, so repititive samples would appear")
                    self.ind_list=list(range(0,self.kbs_data.shape[0]))
                    
                self.x=copy.deepcopy(self.kbs_data[ind,:])
                self.kbs_usage+=1
            else:
                if self.lbound == None and self.ubound == None:
                     self.x=random.sample(self.x,len(self.x))
                elif len(self.lbound) == 1 and len(self.ubound) == 1:
                    #self.x=np.random.randint(self.lbound, self.ubound,len(self.x))
                    for i in range(len(self.x)):
                        if random.random() < self.indpb:
                            self.x[i]=np.random.randint(self.lbound, self.ubound)
                else:
                    for i in range(len(self.x)):
                        if random.random() < self.indpb:
                            self.x[i]=np.random.randint(self.lbound[i], self.ubound[i])
        except:
            raise Exception ('--error: the kbs block cannot be resolved')
            
        return
        
    def build(self):

        """
        This function builds a SA chain and execute it
        """
        random.seed(10)
        
        # initialize x0 
        if self.inp.sa_dict['initstate'][0] == None:
            if self.lbound == None and self.lbound == None:
                self.x=random.sample(range(0,self.inp.gen_dict['xsize'][0]),self.inp.gen_dict['xsize'][0])
            elif len(self.lbound) == 1 and len(self.ubound) == 1:
                self.x=np.random.randint(self.lbound, self.ubound, self.inp.gen_dict['xsize'][0])
            else:
                self.x=np.random.randint(self.lbound, self.ubound)
        else:
            self.x=self.inp.sa_dict['initstate'][0]
        
        # initialize Temperature
        T = self.Tmax
        
        #---------------------------------------------------------------------
        #SA is programmed to minimize but we are maximizing rewards in neorl 
        #so we will be minimizing energy= -reward, which is same as maximizing reward
        #---------------------------------------------------------------------
        # initialize energy 
        energy = -self.fit(self.x)[0] #use[0] as fit returns tuple for GA/DEAP
        # set x/xbest and energy/energy_best to x0, energy0
        x_prev = copy.deepcopy(self.x)
        x_best = copy.deepcopy(self.x)
        energy_prev = energy
        energy_best = energy
        
        # decide the movment mode 
        if self.swap == 'singleswap':
            self.move = self.singleswap
        elif self.swap == 'dualswap':
            self.move = self.dualswap
        elif self.swap == 'quadswap':
            self.move = self.quadswap
        elif self.swap == 'fullswap':
            self.move = self.fullswap
        elif self.swap == 'kbs':
            self.move = self.kbs
        else:
            raise Exception ('--error: the swap mode for SA is not defined, use singleswap, dualswap, quadswap, fullswap, or kbs')
        
        trials, accepts, improves = 0, 0, 0        
        # Attempt moves to new states
        step=0
        fit_lst=[]
        
        #---------------------------------------------
        # Start SA chain 
        #--------------------------------------------
        while step <= self.steps:

            step += 1
            
            # Decide cooling schedule
            if self.cooling=='fast':
                Tfac = -np.log(float(self.Tmax) / self.Tmin)
                T = self.Tmax * np.exp( Tfac * step / self.steps)
            elif self.cooling=='boltzmann':
                T = self.Tmax / np.log(step + 1)
            elif self.cooling=='cauchy':
                T = self.Tmax / (step + 1)
            else:
                raise Exception ('--error: unknown cooling mode is entered, fast, boltzmann, or cauchy are ONLY allowed')
            
            # Perturb input
            self.move()
            
            #---------------------------------------------------------------------
            #SA is programmed to minimize but we are maximizing rewards in neorl 
            #so we will be minimizing energy= -reward, which is same as maximizing reward
            #---------------------------------------------------------------------
            reward=self.fit(self.x)[0] #use[0] as the fit function returns tuple for GA/DEAP
            fit_lst.append(reward)
            energy =-reward  
            
            dE = energy - energy_prev
            trials += 1
            
            #-----------------------------------
            # Accept/Reject
            #-----------------------------------
            alpha=random.random()
            if dE > 0.0 and np.exp(-dE/T) < alpha:
                # Restore previous state if no improvment and accpetance criterion is not satisfied
                self.x = copy.deepcopy(x_prev)
                energy = energy_prev
            else:
                # Accept new state and compare to best state
                accepts += 1
                if dE < 0.0:
                    improves += 1
                x_prev = copy.deepcopy(self.x)
                energy_prev = energy
                if energy < energy_best:
                    x_best = copy.deepcopy(self.x)
                    energy_best = energy
            
            #-----------------------------------
            # Logging and progress check
            #-----------------------------------
            if step % self.check_freq == 0 or step==self.steps:
                
                #print('step=', step, 'kbs=', self.kbs_usage, 'real_good=', self.real_good)
                
                accept_rate=np.round(accepts/trials*100,2)
                improve_rate=np.round(improves/trials*100,2)
                
                #------------
                # plot progress 
                #------------
                self.callback.plot_progress('Annealing Step')
                
                #------------
                # print summary 
                #------------
                print('*****************************************************')
                print('Annealing step={}/{}'.format(step,self.steps))
                print('Current Temperature/Max Temperature: {}/{}'.format(int(T), self.Tmax))
                print('Accept Rate: {}%'.format(accept_rate))
                print('Improvement Rate: {}% '.format(improve_rate))
                print('Best Reward So Far: {}'.format(-np.round(energy_best,3)))
                print('Best Solution So Far: {}'.format(x_best))
                print('*****************************************************')
                
                trials, accepts, improves = 0, 0, 0
                fit_lst=[]


#    This file is part of NEORL.

#    Copyright (c) 2021 Exelon Corporation and MIT Nuclear Science and Engineering
#    NEORL is free software: you can redistribute it and/or modify
#    it under the terms of the MIT LICENSE

#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#    SOFTWARE.

#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Feb 25 10:29:00 2020

@author: majdi
"""


import matplotlib
matplotlib.use('Agg')
import warnings
warnings.filterwarnings("ignore")

#import os, sys
#print(os.getcwd())

#rl_path = os.path.dirname(os.path.realpath(__file__))
#sys.path.append(os.path.join(rl_path,'baselines'))
#print(sys.path)

#print(rl_path)


# External dependencies
import gym
from neorl.rl.baselines.deepq.policies import MlpPolicy
from neorl.rl.baselines.deepq.dqn import DQN
from neorl.utils.test_policy import evaluate_policy
# import input parameters from the user 
from neorl.parsers.PARSER import InputChecker


class DQNAgent(InputChecker):
    def __init__ (self, inp, callback):
        """
        Input: 
            inp: is a dictionary of validated user input {"ncores": 8, "env": 6x6, ...}
            callback: a class of callback built from stable-baselines to allow intervening during training 
                      to process data and save models
        """
        # self.inp.gen_dict ---> GENERAL CARD: env, env_data, nactions, ...etc
        # self.inp.dqn_dict ---> DQN CARD 
        self.inp=inp    # the full user input dictionary
        self.callback=callback 
        self.mode=self.inp.dqn_dict['mode'][0]
        self.log_dir=self.inp.gen_dict['log_dir']
        self.env = gym.make(self.inp.gen_dict['env'][0], casename=self.inp.dqn_dict['casename'][0], 
                            log_dir=self.log_dir, exepath=self.inp.gen_dict['exepath'][0], env_data=self.inp.gen_dict['env_data'][0], env_seed=1)
        
    def build (self):
        """
        This function builds the DQN agent (ONLY 1 Env/Core is supported) based on the selected mode and runs it according to:
            1- Initializes the env
            2- If mode train is selected, train the model from scratch, learn, and save.
            3- If mode continue is selected, provide a path for pretrained model, load the model, learn, and save.
            4- If mode test is selected, provide a path fror pretrained model, load the model and test. 
        """
        #tensorboard activation (if used)
        #to view tensorboard type
        #tensorboard --logdir=./log_dir/{self.casename}_tensorlog 
        if self.inp.dqn_dict['tensorboard'][0]:
            tensorboard_log=self.log_dir+'{}_tensorlog'.format(self.inp.dqn_dict['casename'][0])
        else:
            tensorboard_log=None
            
        
        if self.mode == 'train':
            # Train from scratch, initialize the model and then learn, and save the last model.
            # Callbacks are used if provided 
            model = DQN(MlpPolicy, self.env,
                        gamma=self.inp.dqn_dict['gamma'][0], 
                        learning_rate=self.inp.dqn_dict['learning_rate'][0], 
                        buffer_size=self.inp.dqn_dict['buffer_size'][0], 
                        exploration_fraction=self.inp.dqn_dict['exploration_fraction'][0], 
                        eps_final=self.inp.dqn_dict['eps_final'][0], 
                        learning_starts=self.inp.dqn_dict['learning_starts'][0], 
                        batch_size=self.inp.dqn_dict['batch_size'][0], 
                        target_network_update_freq=self.inp.dqn_dict['target_network_update_freq'][0],
                        eps_init=self.inp.dqn_dict['eps_init'][0],
                        train_freq=self.inp.dqn_dict['train_freq'][0],
                        prioritized_replay=self.inp.dqn_dict['prioritized_replay'][0],
                        verbose=2, seed=1)
            model.learn(total_timesteps=self.inp.dqn_dict['time_steps'][0], callback=self.callback)
            model.save(self.log_dir+self.inp.dqn_dict['casename'][0]+'_lastmodel.pkl')
        
        if self.mode=='continue':
            # load, contine learning, and save last model
            model = DQN.load(self.inp.dqn_dict['model_load_path'][0], env=self.env,
                        gamma=self.inp.dqn_dict['gamma'][0], 
                        learning_rate=self.inp.dqn_dict['learning_rate'][0], 
                        buffer_size=self.inp.dqn_dict['buffer_size'][0], 
                        exploration_fraction=self.inp.dqn_dict['exploration_fraction'][0], 
                        eps_final=self.inp.dqn_dict['eps_final'][0], 
                        learning_starts=self.inp.dqn_dict['learning_starts'][0], 
                        batch_size=self.inp.dqn_dict['batch_size'][0], 
                        target_network_update_freq=self.inp.dqn_dict['target_network_update_freq'][0],
                        eps_init=self.inp.dqn_dict['eps_init'][0],
                        train_freq=self.inp.dqn_dict['train_freq'][0],
                        prioritized_replay=self.inp.dqn_dict['prioritized_replay'][0],
                        verbose=2, seed=1)

            model.learn(total_timesteps=self.inp.dqn_dict['time_steps'][0], callback=self.callback)
            model.save(self.log_dir+self.inp.dqn_dict['casename'][0]+'_lastmodel.pkl')
            
        if self.mode=='test':
            # load and test the agent. Env is recreated since test mode only works in single core
            print('debug: dqn is running in test mode, single core is used to test the policy')
            env = gym.make(self.inp.gen_dict['env'][0], casename=self.inp.dqn_dict['casename'][0],  log_dir=self.log_dir, 
                           exepath=self.inp.gen_dict['exepath'][0], env_data=self.inp.gen_dict['env_data'][0], env_seed=1)
            model = DQN.load(self.inp.dqn_dict['model_load_path'][0])
            evaluate_policy(model, env, log_dir=self.log_dir+'dqn', 
                            n_eval_episodes=self.inp.dqn_dict["n_eval_episodes"][0], render=self.inp.dqn_dict["render"][0])            
    
            
            
# if __name__ =='__main__':
#      inp=InputParam()
#      dqn=DQNAgent(inp)
#      dqn.build()



import numpy as np


class Buffer(object):
    def __init__(self, env, n_steps, size=50000):
        """
        A buffer for observations, actions, rewards, mu's, states, masks and dones values

        :param env: (Gym environment) The environment to learn from
        :param n_steps: (int) The number of steps to run for each environment
        :param size: (int) The buffer size in number of steps
        """
        self.n_env = env.num_envs
        self.n_steps = n_steps
        self.n_batch = self.n_env * self.n_steps
        # Each loc contains n_env * n_steps frames, thus total buffer is n_env * size frames
        self.size = size // self.n_steps

        if len(env.observation_space.shape) > 1:
            self.raw_pixels = True
            self.height, self.width, self.n_channels = env.observation_space.shape
            self.obs_dtype = np.uint8
        else:
            self.raw_pixels = False
            if len(env.observation_space.shape) == 1:
                self.obs_dim = env.observation_space.shape[-1]
            else:
                self.obs_dim = 1
            self.obs_dtype = np.float32

        # Memory
        self.enc_obs = None
        self.actions = None
        self.rewards = None
        self.mus = None
        self.dones = None
        self.masks = None

        # Size indexes
        self.next_idx = 0
        self.num_in_buffer = 0

    def has_atleast(self, frames):
        """
        Check to see if the buffer has at least the asked number of frames

        :param frames: (int) The number of frames checked
        :return: (bool) number of frames in buffer >= number asked
        """
        # Frames per env, so total (n_env * frames) Frames needed
        # Each buffer loc has n_env * n_steps frames
        return self.num_in_buffer >= (frames // self.n_steps)

    def can_sample(self):
        """
        Check if the buffer has at least one frame

        :return: (bool) if the buffer has at least one frame
        """
        return self.num_in_buffer > 0

    def decode(self, enc_obs):
        """
        Get the stacked frames of an observation

        :param enc_obs: ([float]) the encoded observation
        :return: ([float]) the decoded observation
        """
        # enc_obs has shape [n_envs, n_steps + 1, nh, nw, nc]
        # dones has shape [n_envs, n_steps, nh, nw, nc]
        # returns stacked obs of shape [n_env, (n_steps + 1), nh, nw, nc]
        n_env, n_steps = self.n_env, self.n_steps
        if self.raw_pixels:
            obs_dim = [self.height, self.width, self.n_channels]
        else:
            obs_dim = [self.obs_dim]

        obs = np.zeros([1, n_steps + 1, n_env] + obs_dim, dtype=self.obs_dtype)
        # [n_steps + nstack, n_env, nh, nw, nc]
        x_var = np.reshape(enc_obs, [n_env, n_steps + 1] + obs_dim).swapaxes(1, 0)
        obs[-1, :] = x_var

        if self.raw_pixels:
            obs = obs.transpose((2, 1, 3, 4, 0, 5))
        else:
            obs = obs.transpose((2, 1, 3, 0))
        return np.reshape(obs, [n_env, (n_steps + 1)] + obs_dim[:-1] + [obs_dim[-1]])

    def put(self, enc_obs, actions, rewards, mus, dones, masks):
        """
        Adds a frame to the buffer

        :param enc_obs: ([float]) the encoded observation
        :param actions: ([float]) the actions
        :param rewards: ([float]) the rewards
        :param mus: ([float]) the policy probability for the actions
        :param dones: ([bool])
        :param masks: ([bool])
        """
        # enc_obs [n_env, (n_steps + n_stack), nh, nw, nc]
        # actions, rewards, dones [n_env, n_steps]
        # mus [n_env, n_steps, n_act]

        if self.enc_obs is None:
            self.enc_obs = np.empty([self.size] + list(enc_obs.shape), dtype=self.obs_dtype)
            self.actions = np.empty([self.size] + list(actions.shape), dtype=np.int32)
            self.rewards = np.empty([self.size] + list(rewards.shape), dtype=np.float32)
            self.mus = np.empty([self.size] + list(mus.shape), dtype=np.float32)
            self.dones = np.empty([self.size] + list(dones.shape), dtype=np.bool)
            self.masks = np.empty([self.size] + list(masks.shape), dtype=np.bool)

        self.enc_obs[self.next_idx] = enc_obs
        self.actions[self.next_idx] = actions
        self.rewards[self.next_idx] = rewards
        self.mus[self.next_idx] = mus
        self.dones[self.next_idx] = dones
        self.masks[self.next_idx] = masks

        self.next_idx = (self.next_idx + 1) % self.size
        self.num_in_buffer = min(self.size, self.num_in_buffer + 1)

    def take(self, arr, idx, envx):
        """
        Reads a frame from a list and index for the asked environment ids

        :param arr: (np.ndarray) the array that is read
        :param idx: ([int]) the idx that are read
        :param envx: ([int]) the idx for the environments
        :return: ([float]) the askes frames from the list
        """
        n_env = self.n_env
        out = np.empty([n_env] + list(arr.shape[2:]), dtype=arr.dtype)
        for i in range(n_env):
            out[i] = arr[idx[i], envx[i]]
        return out

    def get(self):
        """
        randomly read a frame from the buffer

        :return: ([float], [float], [float], [float], [bool], [float])
                 observations, actions, rewards, mus, dones, maskes
        """
        # returns
        # obs [n_env, (n_steps + 1), nh, nw, n_stack*nc]
        # actions, rewards, dones [n_env, n_steps]
        # mus [n_env, n_steps, n_act]
        n_env = self.n_env
        assert self.can_sample()

        # Sample exactly one id per env. If you sample across envs, then higher correlation in samples from same env.
        idx = np.random.randint(0, self.num_in_buffer, n_env)
        envx = np.arange(n_env)

        dones = self.take(self.dones, idx, envx)
        enc_obs = self.take(self.enc_obs, idx, envx)
        obs = self.decode(enc_obs)
        actions = self.take(self.actions, idx, envx)
        rewards = self.take(self.rewards, idx, envx)
        mus = self.take(self.mus, idx, envx)
        masks = self.take(self.masks, idx, envx)
        return obs, actions, rewards, mus, dones, masks



import mpi4py
import tensorflow as tf
import numpy as np

import shared.tf_util as tf_util


class RunningMeanStd(object):
    def __init__(self, epsilon=1e-2, shape=()):
        """
        calulates the running mean and std of a data stream
        https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Parallel_algorithm

        :param epsilon: (float) helps with arithmetic issues
        :param shape: (tuple) the shape of the data stream's output
        """
        self._sum = tf.get_variable(
            dtype=tf.float64,
            shape=shape,
            initializer=tf.constant_initializer(0.0),
            name="runningsum", trainable=False)
        self._sumsq = tf.get_variable(
            dtype=tf.float64,
            shape=shape,
            initializer=tf.constant_initializer(epsilon),
            name="runningsumsq", trainable=False)
        self._count = tf.get_variable(
            dtype=tf.float64,
            shape=(),
            initializer=tf.constant_initializer(epsilon),
            name="count", trainable=False)
        self.shape = shape

        self.mean = tf.cast(self._sum / self._count, tf.float32)
        self.std = tf.sqrt(tf.maximum(tf.cast(self._sumsq / self._count, tf.float32) - tf.square(self.mean),
                                      1e-2))

        newsum = tf.placeholder(shape=self.shape, dtype=tf.float64, name='sum')
        newsumsq = tf.placeholder(shape=self.shape, dtype=tf.float64, name='var')
        newcount = tf.placeholder(shape=[], dtype=tf.float64, name='count')
        self.incfiltparams = tf_util.function([newsum, newsumsq, newcount], [],
                                              updates=[tf.assign_add(self._sum, newsum),
                                                       tf.assign_add(self._sumsq, newsumsq),
                                                       tf.assign_add(self._count, newcount)])

    def update(self, data):
        """
        update the running mean and std

        :param data: (np.ndarray) the data
        """
        data = data.astype('float64')
        data_size = int(np.prod(self.shape))
        totalvec = np.zeros(data_size * 2 + 1, 'float64')
        addvec = np.concatenate([data.sum(axis=0).ravel(), np.square(data).sum(axis=0).ravel(),
                                 np.array([len(data)], dtype='float64')])
        mpi4py.MPI.COMM_WORLD.Allreduce(addvec, totalvec, op=mpi4py.MPI.SUM)
        self.incfiltparams(totalvec[0: data_size].reshape(self.shape),
                           totalvec[data_size: 2 * data_size].reshape(self.shape), totalvec[2 * data_size])


@tf_util.in_session
def test_dist():
    """
    test the running mean std
    """
    np.random.seed(0)
    p_1, p_2, p_3 = (np.random.randn(3, 1), np.random.randn(4, 1), np.random.randn(5, 1))
    q_1, q_2, q_3 = (np.random.randn(6, 1), np.random.randn(7, 1), np.random.randn(8, 1))

    comm = mpi4py.MPI.COMM_WORLD
    assert comm.Get_size() == 2
    if comm.Get_rank() == 0:
        x_1, x_2, x_3 = p_1, p_2, p_3
    elif comm.Get_rank() == 1:
        x_1, x_2, x_3 = q_1, q_2, q_3
    else:
        assert False

    rms = RunningMeanStd(epsilon=0.0, shape=(1,))
    tf_util.initialize()

    rms.update(x_1)
    rms.update(x_2)
    rms.update(x_3)

    bigvec = np.concatenate([p_1, p_2, p_3, q_1, q_2, q_3])

    def checkallclose(var_1, var_2):
        print(var_1, var_2)
        return np.allclose(var_1, var_2)

    assert checkallclose(
        bigvec.mean(axis=0),
        rms.mean.eval(),
    )
    assert checkallclose(
        bigvec.std(axis=0),
        rms.std.eval(),
    )


if __name__ == "__main__":
    # Run with mpirun -np 2 python <filename>
    test_dist()



#    This file is part of NEORL.

#    Copyright (c) 2021 Exelon Corporation and MIT Nuclear Science and Engineering
#    NEORL is free software: you can redistribute it and/or modify
#    it under the terms of the MIT LICENSE

#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#    SOFTWARE.

# -*- coding: utf-8 -*-
"""
Created on Mon Jun 15 19:37:04 2020

@author: Majdi
"""

import random
import numpy as np
from collections import defaultdict
import copy
import time
import joblib
import multiprocessing
import multiprocessing.pool
from neorl.utils.seeding import set_neorl_seed
class NoDaemonProcess(multiprocessing.Process):
    # make 'daemon' attribute always return False
    def _get_daemon(self):
        return False
    def _set_daemon(self, value):
        pass
    daemon = property(_get_daemon, _set_daemon)

# We sub-class multiprocessing.pool.Pool instead of multiprocessing.Pool
# because the latter is only a wrapper function, not a proper class.
class MyPool(multiprocessing.pool.Pool):
    Process = NoDaemonProcess

class ESMod:
    def __init__ (self, bounds, fit, mu, lambda_, ncores=1, indpb=0.1, cxpb=0.6, mutpb=0.3, smin=0.01, smax=0.5, seed=None):  
        """
        Parallel ES:
        A module for constructing evolution strategy (ES) with parallelization in evaluating the population
        Inputs:
            -bounds (dict): input paramter lower/upper bounds in dictionary form
            -fit (function): fitness function 
            -mu (int): number of individuals to survive for next generation mu < lambda_
            -lambda_ (int): total size of population
            -ncores (int): parallel cores
            -cxpb (float): population crossover probablity for ES
            -mutpb (float): population mutation probablity for ES 
            -indpb (float): independent probability for attribute mutation (ONLY used for continuous attributes)
            -smin (float): minimum bound for strategy vector (fix it to 0.01)
            -smin (float): max bound for strategy vector (fix it to 0.5)
        """
        set_neorl_seed(seed)
        self.bounds=bounds
        self.fit=fit
        self.ncores=ncores
        self.smin=smin
        self.smax=smax
        self.cxpb=cxpb
        self.mutpb=mutpb
        self.indpb=indpb
        self.mu=mu
        self.lambda_=lambda_

        assert self.mu <= self.lambda_, "mu (selected population) must be less than lambda (full population)"
        assert (self.cxpb + self.mutpb) <= 1.0, "The sum of the cxpb and mutpb must be smaller or equal to 1.0"
        assert self.ncores >=1, "Number of cores must be more than or equal 1"

        self.lb=[]
        self.ub=[]
        self.datatype=[]
        
        #infer variable types 
        self.datatype = np.array([bounds[item][0] for item in bounds])
        self.lb = np.array([self.bounds[item][1] for item in self.bounds])
        self.ub = np.array([self.bounds[item][2] for item in self.bounds])
            
    def GenES(self, bounds):
        """
        Individual generator
        Inputs:
            -bounds (dict): input paramter lower/upper bounds in dictionary form
        Returns 
            -ind (list): an individual vector with values sampled from bounds
            -strategy (list): the strategy vector with values between smin and smax
        """
        size = len(bounds.keys()) # size of individual 
        content=[]
        for key in bounds:
            if bounds[key][0] == 'int':
                content.append(random.randint(bounds[key][1], bounds[key][2]))
            elif bounds[key][0] == 'float':
                content.append(random.uniform(bounds[key][1], bounds[key][2]))
            elif bounds[key][0] == 'grid':
                content.append(random.sample(bounds[key][1],1)[0])
            else:
                raise Exception ('unknown data type is given, either int, float, or grid are allowed for parameter bounds')   
        ind=list(content)
        strategy = [random.uniform(self.smin,self.smax) for _ in range(size)]
        return ind, strategy

    def init_pop(self, warmup=100, x_known=None):
        """
        Population intializer 
        Inputs:
            -warmup (int): number of individuals to create and evaluate initially
        Returns 
            -pop (dict): initial population in a dictionary form, looks like this
            
        """
        #initialize the population and strategy and run them in parallel (these samples will be used to initialize the memory)
        pop=defaultdict(list)
        # dict key runs from 0 to warmup-1
        # index 0: individual, index 1: strategy, index 2: fitness 
        # Example: 
        """
        pop={key: [ind, strategy, fitness]}
        pop={0: [[1,2,3,4,5], [0.1,0.2,0.3,0.4,0.5], 1.2], 
             ... 
             99: [[1.1,2.1,3.1,4.1,5.1], [0.1,0.2,0.3,0.4,0.5], 5.2]}
        """
        for i in range (warmup):
            caseid='es_gen{}_ind{}'.format(0,i+1)  #caseid are only for logging purposes to distinguish sample source
            data=self.GenES(self.bounds)
            if x_known:
                pop[i].append(x_known[i])
            else:
                pop[i].append(data[0])
            pop[i].append(data[1])
        
        if self.ncores > 1:  #evaluate warmup in parallel
            core_list=[]
            for key in pop:
                caseid='es_gen{}_ind{}'.format(0,key+1) 
                core_list.append([pop[key][0],caseid])
           
            p=MyPool(self.ncores)
            fitness = p.map(self.gen_object, core_list)
            p.close(); p.join()
            #with joblib.Parallel(n_jobs=self.ncores) as parallel:
            #    fitness=parallel(joblib.delayed(self.gen_object)(item) for item in core_list)
            
            [pop[ind].append(fitness[ind]) for ind in range(len(pop))]
        
        else: #evaluate warmup in series
            for key in pop:
                caseid='es_gen{}_ind{}'.format(0,key+1)
                fitness=self.fit(pop[key][0])
                pop[key].append(fitness)
        
        return pop  #return final pop dictionary with ind, strategy, and fitness

    def ensure_bounds(self, vec): # bounds check

        vec_new = []

        for i, (key, val) in enumerate(self.bounds.items()):
            # less than minimum 
            if vec[i] < self.bounds[key][1]:
                vec_new.append(self.bounds[key][1])
            # more than maximum
            if vec[i] > self.bounds[key][2]:
                vec_new.append(self.bounds[key][2])
            # fine
            if self.bounds[key][1] <= vec[i] <= self.bounds[key][2]:
                vec_new.append(vec[i])
        
        return vec_new

    def ensure_discrete(self, ind):
        
        ind=self.ensure_bounds(ind)
        for i in range(len(ind)):
            if self.datatype[i] == 'int':
                ind[i] = int(ind[i])
        return ind

    def gen_object(self, inp):
        """
        This is a worker for multiprocess Pool
        Inputs:
            inp (list of lists): contains data for each core [[ind1, caseid1], ...,  [indN, caseidN]]
        Returns:
            fitness value (float)
        """
        return self.fit(inp[0])

    def select(self, pop, k=1):
        """
        Select function sorts the population from max to min based on fitness and select k best
        Inputs:
            pop (dict): population in dictionary structure
            k (int): top k individuals are selected
        Returns:
            best_dict (dict): the new ordered dictionary with top k selected 
        """
        
        pop=list(pop.items())
        pop.sort(key=lambda e: e[1][2], reverse=True)
        sorted_dict=dict(pop[:k])
        
        #This block creates a new dict where keys are reset to 0 ... k in order to avoid unordered keys after sort
        best_dict=defaultdict(list)
        index=0
        for key in sorted_dict:
            best_dict[index].append(sorted_dict[key][0])
            best_dict[index].append(sorted_dict[key][1])
            best_dict[index].append(sorted_dict[key][2])
            index+=1
        
        sorted_dict.clear()
        return best_dict

    def cx(self, ind1, ind2, strat1, strat2):
        """Executes a classical two points crossover on both the individuals and their
        strategy. The individuals /strategies should be a list. The crossover points for the
        individual and the strategy are the same.
        
        Inputs:
            -ind1 (list): The first individual participating in the crossover.
            -ind2 (list): The second individual participating in the crossover.
            -strat1 (list): The first evolution strategy participating in the crossover.
            -strat2 (list): The second evolution strategy participating in the crossover.
        Returns:
            The new ind1, ind2, strat1, strat2 after crossover in list form
        """
        size = min(len(ind1), len(ind2))
    
        pt1 = random.randint(1, size)
        pt2 = random.randint(1, size - 1)
        if pt2 >= pt1:
            pt2 += 1
        else:  # Swap the two cx points
            pt1, pt2 = pt2, pt1
    
        ind1[pt1:pt2], ind2[pt1:pt2] = ind2[pt1:pt2], ind1[pt1:pt2]
        strat1[pt1:pt2], strat2[pt1:pt2] = strat2[pt1:pt2], strat1[pt1:pt2]
        
        return ind1, ind2, strat1, strat2


    def mutES(self, ind, strat):
        """Mutate an evolution strategy according to mixed Discrete/Continuous mutation rules
        attribute as described in [Li2013].
        The function mutates discrete/float variables according to their type as indicated in self.bounds
        .. Li, Rui, et al. "Mixed integer evolution strategies for parameter optimization." 
           Evolutionary computation 21.1 (2013): 29-64.
        Inputs:
            -ind (list): individual to be mutated.
            -strat (list): individual strategy to be mutated.
        Returns: 
            -ind (list): new individual after mutatation
            -strat (list): individual strategy after mutatation       

        """
        # Infer the datatype, lower/upper bounds from self.bounds for flexible usage 
        lb=[]; ub=[]; datatype=[]
        for key in self.bounds:
            datatype.append(self.bounds[key][0])
            lb.append(self.bounds[key][1])
            ub.append(self.bounds[key][2])
            
        size = len(ind)
        tau=1/np.sqrt(2*size)
        tau_prime=1/np.sqrt(2*np.sqrt(size))
        
        ind=self.ensure_bounds(ind)  #keep it for choice.remove(x) to work
        
        for i in range(size):
            #--------------------------
            # Discrete ES Mutation 
            #--------------------------
            if datatype[i] == 'int':
                norm=random.gauss(0,1)
                # modify the ind strategy
                strat[i] = 1/(1+(1-strat[i])/strat[i]*np.exp(-tau*norm-tau_prime*random.gauss(0,1)))
                #make a transformation of strategy to ensure it is between smin,smax 
                y=(strat[i]-self.smin)/(self.smax-self.smin)
                if np.floor(y) % 2 == 0:
                    y_prime=np.abs(y-np.floor(y))
                else:
                    y_prime=1-np.abs(y-np.floor(y))
                strat[i] = self.smin + (self.smax-self.smin) * y_prime
                
                # check if this attribute is mutated based on the updated strategy
                if random.random() < strat[i]:
                    
                    if int(lb[i]) == int(ub[i]):
                        ind[i] = int(lb[i])
                    else:
                        # make a list of possiblities after excluding the current value to enforce mutation
                        choices=list(range(lb[i],ub[i]+1))
                        choices.remove(ind[i])
                        # randint is NOT used here since it could re-draw the same integer value, choice is used instead
                        ind[i] = random.choice(choices)
                else:
                    ind[i]=int(ind[i])
            
            #--------------------------
            # Continuous ES Mutation 
            #--------------------------
            elif datatype[i] == 'float':
                norm=random.gauss(0,1)
                strat[i] *= np.exp(tau*norm + tau_prime * random.gauss(0, 1)) #normal mutation of strategy
                ind[i] += strat[i] * random.gauss(0, 1) # update the individual position
                
                #check the new individual falls within lower/upper boundaries
                if ind[i] < lb[i]:
                    ind[i] = lb[i]
                if ind[i] > ub[i]:
                    ind[i] = ub[i]
            
            else:
                raise Exception ('ES mutation strategy works with either int/float datatypes, the type provided cannot be interpreted')
            
        ind=self.ensure_bounds(ind)
        strat=list(np.clip(strat, self.smin, self.smax))
            
        return ind, strat

    def GenOffspring(self, pop):
        """
        
        This function generates the offspring by applying crossover, mutation **or** reproduction. 
        The sum of both probabilities self.cxpb and self.mutpb must be in [0,1]
        The reproduction probability is 1 - cxpb - mutpb
        The new offspring goes for fitness evaluation
        
        Inputs:
            pop (dict): population in dictionary structure
        Returns:
            offspring (dict): new modified population in dictionary structure    
        """
        
        
        pop_indices=list(range(0,len(pop)))
        offspring = defaultdict(list)
        for i in range(self.lambda_):
            alpha = random.random()
            #------------------------------
            # Crossover
            #------------------------------
            if alpha < self.cxpb:            
                index1, index2 = random.sample(pop_indices,2)
                ind1, ind2, strat1, strat2 = self.cx(ind1=list(pop[index1][0]),ind2=list(pop[index2][0]), 
                                                     strat1=list(pop[index1][1]),strat2=list(pop[index2][1]))
            
                ind1=self.ensure_bounds(ind1)
                ind2=self.ensure_bounds(ind2)
                ind1=self.ensure_discrete(ind1)  #check discrete variables after crossover
                ind2=self.ensure_discrete(ind2)  #check discrete variables after crossover
                offspring[i].append(ind1)
                offspring[i].append(strat1)
                #print('crossover is done for sample {} between {} and {}'.format(i,index1,index2))
            #------------------------------
            # Mutation
            #------------------------------
            elif alpha < self.cxpb + self.mutpb:  # Apply mutation
                index = random.choice(pop_indices)
                ind, strat=self.mutES(ind=list(pop[index][0]), strat=list(pop[index][1]))
                offspring[i].append(ind)
                offspring[i].append(strat)
                #print('mutation is done for sample {} based on {}'.format(i,index))
            #------------------------------
            # Reproduction from population
            #------------------------------
            else:                         
                index=random.choice(pop_indices)
                pop[index][0]=self.ensure_discrete(pop[index][0])
                offspring[i].append(pop[index][0])
                offspring[i].append(pop[index][1])
                #print('reproduction is done for sample {} based on {}'.format(i,index))

        for item in offspring:
            offspring[item][1]=list(np.clip(offspring[item][1], self.smin, self.smax))
    
        return offspring

    def evolute(self, population, ngen, caseids=None, verbose=0):
        """This is the :math:`(\mu~,~\lambda)` evolutionary algorithm.
    
        Inputs:
            pop (dict): initial population in dictionary structure
            ngeg (int): number of generations to evolute
            caseids (list of strings): for logging purposes, to attach a casename for each indvidual
            verbose (bool): print summary to screen
        Returns:
            population (dict): final population after running all generations (ngen)
        """
        if caseids:
            assert len(caseids) == self.lambda_ * ngen, \
            'caseids length ({}) is not equal to total number of cases to run: lambda_ ({}) * ngen ({})'. format(len(caseids),self.lambda_ ,ngen)
            case_idx=0
        
        # Begin the evolution process
        for gen in range(1, ngen + 1):
            
            # Vary the population and generate new offspring
            offspring = self.GenOffspring(pop=population)
            
            # Evaluate the individuals with an invalid fitness with multiprocessign Pool
            # create and run the Pool
            if self.ncores > 1:
                t0=time.time()
                core_list=[]
                for key in offspring:
                    core_list.append([offspring[key][0],caseids[case_idx]])
                    case_idx+=1
                
                #initialize a pool
                p=MyPool(self.ncores)
                fitness = p.map(self.gen_object, core_list)
                p.close(); p.join()
                #with joblib.Parallel(n_jobs=self.ncores) as parallel:
                #    fitness=parallel(joblib.delayed(self.gen_object)(item) for item in core_list)
                
                [offspring[ind].append(fitness[ind]) for ind in range(len(offspring))]
                
                self.partime=time.time()-t0
                #print('ES:', self.partime)
            else: #serial calcs
                
                for ind in range(len(offspring)):
                    if caseids:
                        fitness=self.fit(offspring[ind][0])
                        case_idx+=1
                    else:
                        fitness=self.fit(offspring[ind][0])
                    offspring[ind].append(fitness)
                
                self.partime=0
                                                
            # Select the next generation population
            population = copy.deepcopy(self.select(pop=offspring, k=self.mu))
            if verbose:
                inds, rwd=[population[i][0] for i in population], [population[i][2] for i in population]
                mean_strategy=[np.mean(population[i][1]) for i in population]
                print('############################################################')
                print('ES step {}/{}, CX={}, MUT={}, MU={}, LAMBDA={}'.format(gen*self.lambda_,ngen*self.lambda_, np.round(self.cxpb,2), np.round(self.mutpb,2), self.mu, self.lambda_))
                print('############################################################')
                print('Statistics for generation {}'.format(gen))
                print('Best Fitness:', np.round(np.max(rwd),2))
                print('Best Individual:', inds[0])
                print('Max Strategy:', np.round(np.max(mean_strategy),3))
                print('Min Strategy:', np.round(np.min(mean_strategy),3))
                print('Average Strategy:', np.round(np.mean(mean_strategy),3))
                print('############################################################')
        return population, self.partime